# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from inspect import cleandoc
import json

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseConfig, NotifyType
from apprise.plugins.workflows import NotifyWorkflows

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    ##################################
    # NotifyWorkflows
    ##################################
    (
        "workflow://",
        {
            # invalid host details (parsing fails very early)
            "instance": None,
        },
    ),
    (
        "workflow://:@/",
        {
            # invalid host details (parsing fails very early)
            "instance": None,
        },
    ),
    (
        "workflow://host/workflow",
        {
            # workflow provided only, no signature
            "instance": TypeError,
        },
    ),
    (
        "workflow://host:443/^(/signature",
        {
            # invalid workflow provided
            "instance": TypeError,
        },
    ),
    (
        "workflow://host:443/workflow1a/signature/?image=no",
        {
            # All tokens provided - we're good
            # Tests case without image defined
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/workflow1b/signature/",
        {
            # support workflows (s added to end)
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/signature/?id=workflow1c",
        {
            # id= to store workflow id
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/signature/?workflow=workflow1d&wrap=yes",
        {
            # workflow= to store workflow id
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/signature/?workflow=workflow1d&wrap=no",
        {
            # workflow= to store workflow id
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/workflow1e/signature/?api-version=2024-01-01",
        {
            # support api-version which is extracted from webhook
            "instance": NotifyWorkflows,
            # Our expected url(privacy=True) startswith() response
            "privacy_url": "workflow://host:443/w...e/s...e/",
        },
    ),
    (
        "workflows://host:443/workflow1b/signature/?ver=2016-06-01",
        {
            # Support ver= (api-version alias)
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/?id=workflow1b&signature=signature",
        {
            # Support signature= (sig= alias)
            "instance": NotifyWorkflows,
            # Our expected url(privacy=True) startswith() response
            "privacy_url": "workflow://host:443/w...b/s...e/",
        },
    ),
    (
        "workflows://host:443/workflow1e/signature/?powerautomate=yes",
        {
            # support power_automate flag
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/workflow1e/signature/?pa=yes&ver=1995-01-01",
        {
            # support power_automate flag with ver flag
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflows://host:443/workflow1e/signature/?pa=yes",
        {
            # support power_automate flag
            "instance": NotifyWorkflows,
        },
    ),
    # Support native URLs
    (
        (
            "https://server.azure.com:443/workflows/643e69f83c8944/"
            "triggers/manual/paths/invoke?"
            "api-version=2016-06-01&sp=%2Ftriggers%2Fmanual%2Frun&"
            "sv=1.0&sig=KODuebWbDGYFr0z0eu"
        ),
        {
            # All tokens provided - we're good
            "instance": NotifyWorkflows,
            # Our expected url(privacy=True) startswith() response
            "privacy_url": "workflow://server.azure.com:443/6...4/K...u/",
        },
    ),
    (
        (
            "https://server.azure.com:443/"
            "powerautomate/automations/direct/"
            "workflows/643e69f83c8944/"
            "triggers/manual/paths/invoke?"
            "api-version=2022-03-01-preview&sp=%2Ftriggers%2Fmanual%2Frun&"
            "sv=1.0&sig=KODuebWbDGYFr0z0eu"
        ),
        {
            # Power-Automate alternative URL - All tokens provided - we're good
            "instance": NotifyWorkflows,
        },
    ),
    (
        "workflow://host:443/workflow2/signature/",
        {
            "instance": NotifyWorkflows,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "workflow://host:443/workflow3/signature/",
        {
            "instance": NotifyWorkflows,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "workflow://host:443/workflow4/signature/",
        {
            "instance": NotifyWorkflows,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_workflows_urls():
    """NotifyWorkflows() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@pytest.fixture
def workflows_url():
    return "workflow://host:443/workflow/signature"


@pytest.fixture
def request_mock(mocker):
    """Prepare requests mock."""
    mock_post = mocker.patch("requests.post")
    mock_post.return_value = requests.Request()
    mock_post.return_value.status_code = requests.codes.ok
    return mock_post


@pytest.fixture
def simple_template(tmpdir):
    template = tmpdir.join("simple.json")
    template.write(
        cleandoc("""
    {
        "type": "message",
        "attachments": [{
            "contentType": "application/vnd.microsoft.card.adaptive",
            "contentUrl": None,
            "content": {
                "$schema":"http://adaptivecards.io/schemas/adaptive-card.json",
                "type": "AdaptiveCard",
                "version": "1.4",
                "msteams": { "width": "full" },
                "body": [
                    {
                        "type": "TextBlock",
                        "text": "**Test**",
                        "style": "heading"
                    },
                ]
            }
        ]
    }
    """)
    )
    return template


def test_plugin_workflows_simple_test(
    request_mock,
    workflows_url,
):
    """
    NotifyWorkflows() simple testing
    """
    # Instantiate our URL
    obj = Apprise.instantiate(workflows_url)
    assert isinstance(obj, NotifyWorkflows)
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert request_mock.called is True
    assert request_mock.call_args_list[0][0][0].startswith(
        "https://host:443/workflows/workflow/triggers/manual/paths/invoke"
    )
    payload = json.loads(request_mock.call_args_list[0][1]["data"])
    assert "NotifyType." not in request_mock.call_args_list[0][1]["data"]
    assert payload == {
        "type": "message",
        "attachments": [
            {
                "contentType": "application/vnd.microsoft.card.adaptive",
                "contentUrl": None,
                "content": {
                    "$schema": "http://adaptivecards.io/schemas/adaptive-card.json",
                    "type": "AdaptiveCard",
                    "version": "1.4",
                    "body": [
                        {
                            "type": "Image",
                            "url": "https://github.com/caronc/apprise/raw/"
                            "master/apprise/assets/themes/default/"
                            "apprise-info-32x32.png",
                            "height": "32px",
                            "altText": NotifyType.INFO.value,
                        },
                        {
                            "type": "TextBlock",
                            # Verify our Title is set
                            "text": "title",
                            "style": "heading",
                            "weight": "Bolder",
                            "size": "Large",
                            "id": "title",
                        },
                        {
                            "type": "TextBlock",
                            # Verify our Body is set
                            "text": "body",
                            "style": "default",
                            "wrap": True,
                            "id": "body",
                        },
                    ],
                    "msteams": {"width": "full"},
                },
            },
        ],
    }

    request_mock.reset_mock()

    # Instantiate our URL
    obj = Apprise.instantiate(f"{workflows_url}?pa=yes")
    assert isinstance(obj, NotifyWorkflows)
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert request_mock.called is True
    assert request_mock.call_args_list[0][0][0].startswith(
        "https://host:443/powerautomate/automations/direct/"
        "workflows/workflow/triggers/manual/paths/invoke"
    )
    payload = json.loads(request_mock.call_args_list[0][1]["data"])
    assert "NotifyType." not in request_mock.call_args_list[0][1]["data"]

    assert payload == {
        "type": "message",
        "attachments": [
            {
                "contentType": "application/vnd.microsoft.card.adaptive",
                "contentUrl": None,
                "content": {
                    "$schema": "http://adaptivecards.io/schemas/"
                    "adaptive-card.json",
                    "type": "AdaptiveCard",
                    "version": "1.4",
                    "body": [
                        {
                            "type": "Image",
                            "url": "https://github.com/caronc/apprise/raw/"
                            "master/apprise/assets/themes/default/"
                            "apprise-info-32x32.png",
                            "height": "32px",
                            "altText": "info",
                        },
                        {
                            "type": "TextBlock",
                            # Verify our Title is set
                            "text": "title",
                            "style": "heading",
                            "weight": "Bolder",
                            "size": "Large",
                            "id": "title",
                        },
                        {
                            "type": "TextBlock",
                            # Verify our Body is set
                            "text": "body",
                            "style": "default",
                            "wrap": True,
                            "id": "body",
                        },
                    ],
                    "msteams": {
                        "width": "full",
                    },
                },
            },
        ],
    }


def test_plugin_workflows_templating_basic_success(
    request_mock, workflows_url, tmpdir
):
    """
    NotifyWorkflows() Templating - success.
    Test cases where URL and JSON is valid.
    """

    template = tmpdir.join("simple.json")
    template.write(
        cleandoc("""
    {
      "@type": "MessageCard",
      "@context": "https://schema.org/extensions",
      "summary": "{{app_id}}",
      "themeColor": "{{app_color}}",
      "sections": [
        {
          "activityImage": null,
          "activityTitle": "{{app_title}}",
          "text": "{{app_body}}"
        }
      ]
    }
    """)
    )

    # Instantiate our URL
    obj = Apprise.instantiate(
        "{url}/?template={template}&{kwargs}".format(
            url=workflows_url,
            template=str(template),
            kwargs=":key1=token&:key2=token",
        )
    )

    assert isinstance(obj, NotifyWorkflows)
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert request_mock.called is True
    assert request_mock.call_args_list[0][0][0].startswith(
        "https://host:443/workflows/workflow/triggers/manual/paths/invoke"
    )

    # Our Posted JSON Object
    posted_json = json.loads(request_mock.call_args_list[0][1]["data"])
    assert "NotifyType." not in request_mock.call_args_list[0][1]["data"]
    assert "summary" in posted_json
    assert posted_json["summary"] == "Apprise"
    assert posted_json["themeColor"] == "#3AA3E3"
    assert posted_json["sections"][0]["activityTitle"] == "title"
    assert posted_json["sections"][0]["text"] == "body"


def test_plugin_workflows_templating_invalid_json(
    request_mock, workflows_url, tmpdir
):
    """
    NotifyWorkflows() Templating - invalid JSON.
    """

    template = tmpdir.join("invalid.json")
    template.write("}")

    # Instantiate our URL
    obj = Apprise.instantiate(
        "{url}/?template={template}&{kwargs}".format(
            url=workflows_url,
            template=str(template),
            kwargs=":key1=token&:key2=token",
        )
    )

    assert isinstance(obj, NotifyWorkflows)
    # We will fail to preform our notifcation because the JSON is bad
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )


def test_plugin_workflows_templating_load_json_failure(
    request_mock, workflows_url, tmpdir
):
    """
    NotifyWorkflows() Templating - template loading failure.
    Test a case where we can not access the file.
    """

    template = tmpdir.join("empty.json")
    template.write("")

    obj = Apprise.instantiate(f"{workflows_url}/?template={template!s}")

    with mock.patch("json.loads", side_effect=OSError):
        # we fail, but this time it's because we couldn't
        # access the cached file contents for reading
        assert (
            obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
            is False
        )


def test_plugin_workflows_templating_target_success(
    request_mock, workflows_url, tmpdir
):
    """
    NotifyWorkflows() Templating - success with target.
    A more complicated example; uses a target.
    """

    template = tmpdir.join("more_complicated_example.json")
    template.write(
        cleandoc("""
    {
      "@type": "MessageCard",
      "@context": "https://schema.org/extensions",
      "summary": "{{app_desc}}",
      "themeColor": "{{app_color}}",
      "sections": [
        {
          "activityImage": null,
          "activityTitle": "{{app_title}}",
          "text": "{{app_body}}"
        }
      ],
     "potentialAction": [{
        "@type": "ActionCard",
        "name": "Add a comment",
        "inputs": [{
            "@type": "TextInput",
            "id": "comment",
            "isMultiline": false,
            "title": "Add a comment here for this task."
        }],
        "actions": [{
            "@type": "HttpPOST",
            "name": "Add Comment",
            "target": "{{ target }}"
        }]
     }]
    }
    """)
    )

    # Instantiate our URL
    obj = Apprise.instantiate(
        "{url}/?template={template}&{kwargs}".format(
            url=workflows_url,
            template=str(template),
            kwargs=":key1=token&:key2=token&:target=http://localhost",
        )
    )

    assert isinstance(obj, NotifyWorkflows)
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert request_mock.called is True
    assert request_mock.call_args_list[0][0][0].startswith(
        "https://host:443/workflows/workflow/triggers/manual/paths/invoke"
    )

    # Our Posted JSON Object
    posted_json = json.loads(request_mock.call_args_list[0][1]["data"])
    assert "NotifyType." not in request_mock.call_args_list[0][1]["data"]
    assert "summary" in posted_json
    assert posted_json["summary"] == "Apprise Notifications"
    assert posted_json["themeColor"] == "#3AA3E3"
    assert posted_json["sections"][0]["activityTitle"] == "title"
    assert posted_json["sections"][0]["text"] == "body"

    # We even parsed our entry out of the URL
    assert (
        posted_json["potentialAction"][0]["actions"][0]["target"]
        == "http://localhost"
    )


def test_workflows_yaml_config_missing_template_filename(
    request_mock, workflows_url, simple_template, tmpdir
):
    """
    NotifyWorkflows() YAML Configuration Entries - Missing template reference.
    """

    config = tmpdir.join("workflow01.yml")
    config.write(
        cleandoc(f"""
    urls:
      - {workflows_url}:
        - tag: 'workflow'
          template: {simple_template!s}.missing
          :name: 'Template.Missing'
          :body: 'test body'
          :title: 'test title'
    """)
    )

    # Config still loads okay
    cfg = AppriseConfig()
    cfg.add(str(config))
    assert len(cfg) == 1
    assert len(cfg[0]) == 1

    obj = cfg[0][0]
    assert isinstance(obj, NotifyWorkflows)

    # However we can't send notification since the template couldn't be loaded
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )
    assert request_mock.called is False


def test_plugin_workflows_edge_cases():
    """NotifyWorkflows() Edge Cases."""
    # Initializes the plugin with an invalid token
    with pytest.raises(TypeError):
        NotifyWorkflows(workflow="@", signature="@")
    with pytest.raises(TypeError):
        NotifyWorkflows(workflow="", signature="abcd")

    with pytest.raises(TypeError):
        NotifyWorkflows(workflow=None, signature="abcd")
    # Whitespace also acts as an invalid token value
    with pytest.raises(TypeError):
        NotifyWorkflows(workflow="  ", signature="abcd")

    with pytest.raises(TypeError):
        NotifyWorkflows(workflow="abcd", signature=None)
    # Whitespace also acts as an invalid token value
    with pytest.raises(TypeError):
        NotifyWorkflows(workflow="abcd", signature="  ")

    # test case where invalid tokens are specified
    with pytest.raises(TypeError):
        NotifyWorkflows(
            workflow="workflow", signature="signature", tokens="not-a-dict"
        )

    # test case where no tokens are specified
    obj = NotifyWorkflows(workflow="workflow", signature="signature")
    assert isinstance(obj, NotifyWorkflows)


def test_plugin_workflows_azure_webhooks(request_mock):
    """NotifyWorkflows() Azure Webhooks."""
    url = (
        "https://prod-15.uksouth.logic.azure.com:443"
        "/workflows/3XXX5/triggers/manual/paths/invoke"
        "?api-version=2016-06-01&"
        "sp=%2Ftriggers%2Fmanual%2Frun&sv=1.0&sig=iXXXU"
    )

    #
    # Initialize
    #
    obj = Apprise.instantiate(url)
    assert isinstance(obj, NotifyWorkflows)
    assert obj.workflow == "3XXX5"
    assert obj.signature == "iXXXU"
    assert obj.api_version == "2016-06-01"



#
# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import logging

from helpers import AppriseURLTester
import requests

from apprise.plugins.spike import NotifySpike

logging.disable(logging.CRITICAL)

apprise_url_tests = (
    (
        "spike://",
        {
            "instance": TypeError,
        },
    ),
    (
        "spike://invalid-key",
        {
            "instance": TypeError,
        },
    ),
    (
        "spike://1234567890abcdef1234567890abcdef",
        {
            "instance": NotifySpike,
            "privacy_url": "spike://****/",
        },
    ),
    (
        "spike://?token=1234567890abcdef1234567890abcdef",
        {
            "instance": NotifySpike,
            "privacy_url": "spike://****/",
        },
    ),
    (
        "https://api.spike.sh/v1/alerts/1234567890abcdef1234567890abcdef",
        {
            "instance": NotifySpike,
        },
    ),
    (
        "spike://1234567890abcdef1234567890abcdef",
        {
            "instance": NotifySpike,
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
        },
    ),
    (
        "spike://1234567890abcdef1234567890abcdef",
        {
            "instance": NotifySpike,
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "spike://ffffffffffffffffffffffffffffffff",
        {
            "instance": NotifySpike,
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_spike_urls():
    AppriseURLTester(tests=apprise_url_tests).run_all()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment
from apprise.plugins.pushbullet import NotifyPushBullet

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    (
        "pbul://",
        {
            "instance": TypeError,
        },
    ),
    (
        "pbul://:@/",
        {
            "instance": TypeError,
        },
    ),
    # APIkey
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # APIkey; but support attachment response
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "file_name": "cat.jpeg",
                "file_type": "image/jpeg",
                "file_url": "http://file_url",
                "upload_url": "http://upload_url",
            },
        },
    ),
    # APIkey; attachment testing that isn't an image type
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "file_name": "test.pdf",
                "file_type": "application/pdf",
                "file_url": "http://file_url",
                "upload_url": "http://upload_url",
            },
        },
    ),
    # APIkey; attachment testing were expected entry in payload is missing
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "file_name": "test.pdf",
                "file_type": "application/pdf",
                "file_url": "http://file_url",
                # upload_url missing
            },
            # Our Notification calls associated with attachments will fail:
            "attach_response": False,
        },
    ),
    # API Key + channel
    (
        "pbul://%s/#channel/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + channel (via to=
    (
        "pbul://%s/?to=#channel" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + 2 channels
    (
        "pbul://%s/#channel1/#channel2" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "pbul://a...a/",
            "check_attachments": False,
        },
    ),
    # API Key + device
    (
        "pbul://%s/device/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + 2 devices
    (
        "pbul://%s/device1/device2/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + email
    (
        "pbul://%s/user@example.com/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + 2 emails
    (
        "pbul://%s/user@example.com/abc@def.com/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # API Key + Combo
    (
        "pbul://%s/device/#channel/user@example.com/" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            "check_attachments": False,
        },
    ),
    # ,
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
            "check_attachments": False,
        },
    ),
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
            "check_attachments": False,
        },
    ),
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
            "check_attachments": False,
        },
    ),
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # force a failure
            "response": False,
            "requests_response_code": requests.codes.internal_server_error,
            "check_attachments": False,
        },
    ),
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
            "check_attachments": False,
        },
    ),
    (
        "pbul://%s" % ("a" * 32),
        {
            "instance": NotifyPushBullet,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
            "check_attachments": False,
        },
    ),
)


def test_plugin_pushbullet_urls():
    """NotifyPushBullet() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_pushbullet_attachments(mock_post):
    """NotifyPushBullet() Attachment Checks."""

    # Initialize some generic (but valid) tokens
    access_token = "t" * 32

    # Prepare Mock return object
    response = mock.Mock()
    response.content = dumps(
        {
            "file_name": "cat.jpg",
            "file_type": "image/jpeg",
            "file_url": "https://dl.pushb.com/abc/cat.jpg",
            "upload_url": "https://upload.pushbullet.com/abcd123",
        }
    ).encode("utf-8")
    response.status_code = requests.codes.ok
    mock_post.return_value = response

    # prepare our attachment
    attach = AppriseAttachment(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))

    # Test our markdown
    obj = Apprise.instantiate(f"pbul://{access_token}/?format=markdown")

    # Send a good attachment
    assert obj.notify(body="test", attach=attach) is True

    # Test our call count
    assert mock_post.call_count == 4
    # Image Prep
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.pushbullet.com/v2/upload-request"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://upload.pushbullet.com/abcd123"
    )
    # Message
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://api.pushbullet.com/v2/pushes"
    )
    # Image Send
    assert (
        mock_post.call_args_list[3][0][0]
        == "https://api.pushbullet.com/v2/pushes"
    )

    # Reset our mock object
    mock_post.reset_mock()

    # Add another attachment so we drop into the area of the PushBullet code
    # that sends remaining attachments (if more detected)
    attach.add(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))

    # Send our attachments
    assert obj.notify(body="test", attach=attach) is True

    # Test our call count
    assert mock_post.call_count == 7
    # Image Prep
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://api.pushbullet.com/v2/upload-request"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://upload.pushbullet.com/abcd123"
    )
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://api.pushbullet.com/v2/upload-request"
    )
    assert (
        mock_post.call_args_list[3][0][0]
        == "https://upload.pushbullet.com/abcd123"
    )
    # Message
    assert (
        mock_post.call_args_list[4][0][0]
        == "https://api.pushbullet.com/v2/pushes"
    )
    # Image Send
    assert (
        mock_post.call_args_list[5][0][0]
        == "https://api.pushbullet.com/v2/pushes"
    )
    assert (
        mock_post.call_args_list[6][0][0]
        == "https://api.pushbullet.com/v2/pushes"
    )

    # Reset our mock object
    mock_post.reset_mock()

    # An invalid attachment will cause a failure
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    attach = AppriseAttachment(path)
    assert obj.notify(body="test", attach=attach) is False

    # Test our call count
    assert mock_post.call_count == 0

    # prepare our attachment
    attach = AppriseAttachment(os.path.join(TEST_VAR_DIR, "apprise-test.gif"))

    # Prepare a bad response
    bad_response = mock.Mock()
    bad_response.content = dumps(
        {
            "file_name": "cat.jpg",
            "file_type": "image/jpeg",
            "file_url": "https://dl.pushb.com/abc/cat.jpg",
            "upload_url": "https://upload.pushbullet.com/abcd123",
        }
    ).encode("utf-8")
    bad_response.status_code = requests.codes.internal_server_error
    bad_response.headers = {}

    # Prepare a bad response
    bad_json_response = mock.Mock()
    bad_json_response.content = b"}"
    bad_json_response.status_code = requests.codes.ok
    bad_json_response.headers = {}

    # Throw an exception on the first call to requests.post()
    for side_effect in (
        requests.RequestException(),
        OSError(),
        [bad_response],
    ):
        mock_post.reset_mock()
        mock_post.side_effect = side_effect

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # Throw an exception on the second call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.reset_mock()
        mock_post.side_effect = [response, side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # Throw an exception on the third call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.reset_mock()
        mock_post.side_effect = [response, response, side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # Throw an exception on the forth call to requests.post()
    for side_effect in (requests.RequestException(), OSError(), bad_response):
        mock_post.reset_mock()
        mock_post.side_effect = [response, response, response, side_effect]

        # We'll fail now because of our error handling
        assert obj.send(body="test", attach=attach) is False

    # Test case where we don't get a valid response back
    mock_post.side_effect = None
    mock_post.return_value = bad_json_response

    # We'll fail because of an invalid json object
    assert obj.send(body="test", attach=attach) is False


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_pushbullet_edge_cases(mock_post, mock_get):
    """NotifyPushBullet() Edge Cases."""

    # Initialize some generic (but valid) tokens
    accesstoken = "a" * 32

    # Support strings
    recipients = "#chan1,#chan2,device,user@example.com,,,"

    # Prepare Mock
    mock_get.return_value = requests.Request()
    mock_post.return_value = requests.Request()
    mock_post.content = b""
    mock_post.return_value.status_code = requests.codes.ok
    mock_get.return_value.status_code = requests.codes.ok
    mock_get.content = b""

    # Invalid Access Token
    with pytest.raises(TypeError):
        NotifyPushBullet(accesstoken=None)
    with pytest.raises(TypeError):
        NotifyPushBullet(accesstoken="     ")

    obj = NotifyPushBullet(accesstoken=accesstoken, targets=recipients)
    assert isinstance(obj, NotifyPushBullet) is True
    assert len(obj.targets) == 4

    obj = NotifyPushBullet(accesstoken=accesstoken)
    assert isinstance(obj, NotifyPushBullet) is True
    # Default is to send to all devices, so there will be a
    # recipient here
    assert len(obj.targets) == 1

    obj = NotifyPushBullet(accesstoken=accesstoken, targets=set())
    assert isinstance(obj, NotifyPushBullet) is True
    # Default is to send to all devices, so there will be a
    # recipient here
    assert len(obj.targets) == 1



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from json import dumps

# Disable logging for a cleaner testing output
import logging
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise
from apprise.plugins.twilio import NotifyTwilio, TwilioNotificationMethod

logging.disable(logging.CRITICAL)

# Our Testing URLs
apprise_url_tests = (
    (
        "twilio://",
        {
            # No Account SID specified
            "instance": TypeError,
        },
    ),
    (
        "twilio://:@/",
        {
            # invalid Auth token
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}@12345678".format("a" * 32),
        {
            # Just sid provided
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}:{}@_".format("a" * 32, "b" * 32),
        {
            # sid and token provided but invalid from
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}:{}@{}".format("a" * 32, "b" * 32, "3" * 5),
        {
            # using short-code (5 characters) without a target
            # We can still instantiate ourselves with a valid short code
            "instance": NotifyTwilio,
            # Since there are no targets specified we expect a False return on
            # send()
            "notify_response": False,
        },
    ),
    (
        "twilio://AC{}:{}@{}".format("a" * 32, "b" * 32, "3" * 9),
        {
            # sid and token provided and from but invalid from no
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}:{}@{}/123/{}/abcd/w:{}".format(
            "a" * 32, "b" * 32, "3" * 11, "9" * 15, 8 * 11
        ),
        {
            # valid everything but target numbers
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://AC{}:{}@12345/{}".format("a" * 32, "b" * 32, "4" * 11),
        {
            # using short-code (5 characters)
            "instance": NotifyTwilio,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "twilio://...aaaa:b...b@12345",
        },
    ),
    (
        "twilio://AC{}:{}@98765/{}/w:{}/".format(
            "a" * 32, "b" * 32, "4" * 11, "5" * 11
        ),
        {
            # using short-code (5 characters) and 1 twillio address ignored
            # because source phone number can not be a short code
            "instance": NotifyTwilio,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "twilio://...aaaa:b...b@98765",
        },
    ),
    (
        "twilio://AC{}:{}@w:12345/{}/{}".format(
            "a" * 32, "b" * 32, "4" * 11, "5" * 11
        ),
        {
            # Invalid short-code
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}:{}@123456/{}".format("a" * 32, "b" * 32, "4" * 11),
        {
            # using short-code (6 characters)
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://AC{}:{}@{}".format("a" * 32, "b" * 32, "5" * 11),
        {
            # using phone no with no target - we text ourselves in
            # this case
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://AC{}:{}@{}?method=sms".format("a" * 32, "b" * 32, "5" * 11),
        {
            # Specify explicitly notification method sms
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://AC{}:{}@{}?method=mms".format("a" * 32, "b" * 32, "5" * 11),
        {
            # Invalid notification method
            "instance": TypeError,
        },
    ),
    (
        "twilio://AC{}:{}@{}?method=call".format(
            "a" * 32, "b" * 32, "w:" + "5" * 11
        ),
        {
            # Incompatibility between Whatsapp mode and CALL method
            "instance": TypeError,
        },
    ),
    (
        "twilio://_?sid=AC{}&token={}&from={}".format(
            "a" * 32, "b" * 32, "5" * 11
        ),
        {
            # use get args to acomplish the same thing
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://_?sid=AC{}&token={}&from={}&to=w:{}".format(
            "a" * 32, "b" * 32, "5" * 11, "6" * 11
        ),
        {
            # Support whatsapp (w: before number)
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://_?sid=AC{}&token={}&source={}".format(
            "a" * 32, "b" * 32, "5" * 11
        ),
        {
            # use get args to acomplish the same thing (use source instead of
            # from)
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://_?sid=AC{}&token={}&from={}&to={}".format(
            "a" * 32, "b" * 32, "5" * 11, "7" * 13
        ),
        {
            # use to=
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://_?sid=AC{}&token={}&from={}&to={}method=call".format(
            "a" * 32, "b" * 32, "5" * 11, "7" * 13
        ),
        {
            # Specify notification method call
            "instance": NotifyTwilio,
        },
    ),
    (
        "twilio://AC{}:{}@{}".format("a" * 32, "b" * 32, "6" * 11),
        {
            "instance": NotifyTwilio,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "twilio://AC{}:{}@{}".format("a" * 32, "b" * 32, "6" * 11),
        {
            "instance": NotifyTwilio,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
)


def test_plugin_twilio_urls():
    """NotifyTwilio() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.post")
def test_plugin_twilio_auth(mock_post):
    """NotifyTwilio() Auth.

    - account-wide auth token
    - API key and its own auth token
    """

    response = mock.Mock()
    response.content = ""
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # Initialize some generic (but valid) tokens
    account_sid = "AC{}".format("b" * 32)
    apikey = "SK{}".format("b" * 32)
    auth_token = "{}".format("b" * 32)
    source = "+1 (555) 123-3456"
    dest = "+1 (555) 987-6543"
    message_contents = "test"

    # Variation of initialization without API key
    obj = Apprise.instantiate(
        f"twilio://{account_sid}:{auth_token}@{source}/{dest}"
    )
    assert isinstance(obj, NotifyTwilio)
    assert isinstance(obj.url(), str)

    # Send Notification
    assert obj.send(body=message_contents) is True

    # Variation of initialization with API key
    obj = Apprise.instantiate(
        f"twilio://{account_sid}:{auth_token}@{source}/{dest}?apikey={apikey}"
    )
    assert isinstance(obj, NotifyTwilio)
    assert isinstance(obj.url(), str)

    # Send Notification
    assert obj.send(body=message_contents) is True

    # Variation of initialization with method call
    obj = Apprise.instantiate(
        f"twilio://{account_sid}:{auth_token}@{source}/{dest}?method=call"
    )
    assert isinstance(obj, NotifyTwilio)
    assert isinstance(obj.url(), str)

    # Send Notification
    assert obj.send(body=message_contents) is True

    # Validate expected call parameters
    assert mock_post.call_count == 3
    first_call = mock_post.call_args_list[0]
    second_call = mock_post.call_args_list[1]
    third_call = mock_post.call_args_list[2]

    # URL and message parameters are the same for both calls
    assert (
        first_call[0][0]
        == second_call[0][0]
        == (
            "https://api.twilio.com/2010-04-01/Accounts"
            f"/{account_sid}/Messages.json"
        )
    )
    assert (
        first_call[1]["data"]["Body"]
        == second_call[1]["data"]["Body"]
        == message_contents
    )
    assert third_call[0][0] == (
        f"https://api.twilio.com/2010-04-01/Accounts/{account_sid}/Calls.json"
    )
    assert third_call[1]["data"]["Twiml"] == message_contents
    assert (
        first_call[1]["data"]["From"]
        == second_call[1]["data"]["From"]
        == third_call[1]["data"]["From"]
        == "+15551233456"
    )
    assert (
        first_call[1]["data"]["To"]
        == second_call[1]["data"]["To"]
        == third_call[1]["data"]["To"]
        == "+15559876543"
    )

    # Auth differs depending on if API Key is used
    assert first_call[1]["auth"] == (account_sid, auth_token)
    assert second_call[1]["auth"] == (apikey, auth_token)
    assert third_call[1]["auth"] == (account_sid, auth_token)


@mock.patch("requests.post")
def test_plugin_twilio_edge_cases(mock_post):
    """NotifyTwilio() Edge Cases."""

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_post.return_value = response

    # Initialize some generic (but valid) tokens
    account_sid = "AC{}".format("b" * 32)
    auth_token = "{}".format("b" * 32)
    source = "+1 (555) 123-3456"
    whatsapp_source = "w:" + "+1 (555) 123-3456"

    # No account_sid specified
    with pytest.raises(TypeError):
        NotifyTwilio(account_sid=None, auth_token=auth_token, source=source)

    # No auth_token specified
    with pytest.raises(TypeError):
        NotifyTwilio(account_sid=account_sid, auth_token=None, source=source)

    # Source is bad
    with pytest.raises(TypeError):
        NotifyTwilio(account_sid=account_sid, auth_token=auth_token, source="")

    # Incompatibility between mode and method
    with pytest.raises(TypeError):
        NotifyTwilio(
            account_sid=account_sid,
            auth_token=auth_token,
            source=whatsapp_source,
            method=TwilioNotificationMethod.CALL,
        )

    # a error response
    response.status_code = 400
    response.content = dumps(
        {
            "code": 21211,
            "message": "The 'To' number +1234567 is not a valid phone number.",
        }
    )
    mock_post.return_value = response

    # Initialize our object
    obj = NotifyTwilio(
        account_sid=account_sid, auth_token=auth_token, source=source
    )

    # We will fail with the above error code
    assert obj.notify("title", "body", "info") is False



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from inspect import cleandoc
import json

# Disable logging for a cleaner testing output
import logging
from os.path import dirname, getsize, join
from unittest import mock

import pytest
import requests

from apprise import Apprise, AppriseAsset, AttachmentManager
from apprise.apprise_attachment import AppriseAttachment
from apprise.attachment import AttachBase
from apprise.common import ContentLocation

logging.disable(logging.CRITICAL)

TEST_VAR_DIR = join(dirname(__file__), "var")

# Grant access to our Attachment Manager Singleton
A_MGR = AttachmentManager()


def test_apprise_attachment():
    """
    API: AppriseAttachment basic testing

    """

    # Create ourselves an attachment object
    aa = AppriseAttachment()

    # There are no attachents loaded
    assert len(aa) == 0

    # Object can be directly checked as a boolean; response is False
    # when there are no entries loaded
    assert not aa

    # An attachment object using a custom Apprise Asset object
    # Set a cache expiry of 5 minutes (300 seconds)
    aa = AppriseAttachment(asset=AppriseAsset(), cache=300)

    # still no attachments added
    assert len(aa) == 0

    # Add a file by it's path
    path = join(TEST_VAR_DIR, "apprise-test.gif")
    assert aa.add(path)

    # There is now 1 attachment
    assert len(aa) == 1

    # our attachment took on our cache value
    assert aa[0].cache == 300

    # we can test the object as a boolean and get a value of True now
    assert aa

    # Add another entry already in it's AttachBase format
    response = AppriseAttachment.instantiate(path, cache=True)
    assert isinstance(response, AttachBase)
    assert aa.add(response, asset=AppriseAsset())

    # There is now 2 attachments
    assert len(aa) == 2

    # Cache is initialized to True
    assert aa[1].cache is True

    # Reset our object
    aa = AppriseAttachment()

    # We can add by lists as well in a variety of formats
    attachments = (
        path,
        f"file://{path}?name=newfilename.gif?cache=120",
        AppriseAttachment.instantiate(
            f"file://{path}?name=anotherfilename.gif", cache=100
        ),
    )

    # Add them
    assert aa.add(attachments, cache=False)

    # There is now 3 attachments
    assert len(aa) == 3

    # Take on our fixed cache value of False.
    # The last entry will have our set value of 100
    assert aa[0].cache is False
    # Even though we set a value of 120, we take on the value of False because
    # it was forced on the instantiate call
    assert aa[1].cache is False
    assert aa[2].cache == 100

    # We can pop the last element off of the list as well
    attachment = aa.pop()
    assert isinstance(attachment, AttachBase)
    # we can test of the attachment is valid using a boolean check:
    assert attachment
    assert len(aa) == 2
    assert attachment.path == path
    assert attachment.name == "anotherfilename.gif"
    assert attachment.mimetype == "image/gif"

    # elements can also be directly indexed
    assert isinstance(aa[0], AttachBase)
    assert isinstance(aa[1], AttachBase)

    with pytest.raises(IndexError):
        aa[2]

    # We can iterate over attachments too:
    for count, a in enumerate(aa):
        assert isinstance(a, AttachBase)

        # we'll never iterate more then the number of entries in our object
        assert count < len(aa)

    # Get the file-size of our image
    expected_size = getsize(path) * len(aa)

    # verify that's what we get as a result
    assert aa.size() == expected_size

    # Attachments can also be loaded during the instantiation of the
    # AppriseAttachment object
    aa = AppriseAttachment(attachments)

    # There is now 3 attachments
    assert len(aa) == 3

    # Reset our object
    aa.clear()
    assert len(aa) == 0
    assert not aa

    assert aa.add(
        AppriseAttachment.instantiate(
            f"file://{path}?name=andanother.png&cache=Yes"
        )
    )
    assert aa.add(
        AppriseAttachment.instantiate(
            f"file://{path}?name=andanother.png&cache=No"
        )
    )
    AppriseAttachment.instantiate(
        f"file://{path}?name=andanother.png&cache=600"
    )
    assert aa.add(
        AppriseAttachment.instantiate(
            f"file://{path}?name=andanother.png&cache=600"
        )
    )

    assert len(aa) == 3
    assert aa[0].cache is True
    assert aa[1].cache is False
    assert aa[2].cache == 600

    # Negative cache are not allowed
    assert not aa.add(
        AppriseAttachment.instantiate(
            f"file://{path}?name=andanother.png&cache=-600"
        )
    )

    # Invalid cache value
    assert not aa.add(
        AppriseAttachment.instantiate(
            f"file://{path}?name=andanother.png", cache="invalid"
        )
    )

    # No length change
    assert len(aa) == 3

    # Reset our object
    aa.clear()

    # Garbage in produces garbage out
    assert aa.add(None) is False
    assert aa.add(object()) is False
    assert aa.add(42) is False

    # length remains unchanged
    assert len(aa) == 0

    # We can add by lists as well in a variety of formats
    attachments = (
        None,
        object(),
        42,
        "garbage://",
    )

    # Add our attachments
    assert aa.add(attachments) is False

    # length remains unchanged
    assert len(aa) == 0

    # if instantiating attachments from the class, it will throw a TypeError
    # if attachments couldn't be loaded
    with pytest.raises(TypeError):
        AppriseAttachment("garbage://")

    # Load our other attachment types
    aa = AppriseAttachment(location=ContentLocation.LOCAL)

    # Hosted type won't allow us to import files
    aa = AppriseAttachment(location=ContentLocation.HOSTED)
    assert len(aa) == 0

    # Add our attachments defined a the head of this function
    aa.add(attachments)

    # Our length is still zero because we can't import files in
    # a hosted environment
    assert len(aa) == 0

    # Inaccessible type prevents the adding of new stuff
    aa = AppriseAttachment(location=ContentLocation.INACCESSIBLE)
    assert len(aa) == 0

    # Add our attachments defined a the head of this function
    aa.add(attachments)

    # Our length is still zero
    assert len(aa) == 0

    with pytest.raises(TypeError):
        # Invalid location specified
        AppriseAttachment(location="invalid")

    # test cases when file simply doesn't exist
    aa = AppriseAttachment("file://non-existant-file.png")
    # Our length is still 1
    assert len(aa) == 1
    # Our object will still return a True
    assert aa

    # However our indexed entry will not
    assert not aa[0]

    # length will return 0
    assert len(aa[0]) == 0

    # Total length will also return 0
    assert aa.size() == 0


@mock.patch("requests.request")
def test_apprise_attachment_truncate(mock_request):
    """
    API: AppriseAttachment when truncation in place

    """

    # Prepare our response
    response = requests.Request()
    response.status_code = requests.codes.ok

    # Prepare Mock
    mock_request.return_value = response

    # our Apprise Object
    ap_obj = Apprise()

    # Add ourselves an object set to truncate
    ap_obj.add("json://localhost/?method=GET&overflow=truncate")

    # Create ourselves an attachment object
    aa = AppriseAttachment()

    # There are no attachents loaded
    assert len(aa) == 0

    # Object can be directly checked as a boolean; response is False
    # when there are no entries loaded
    assert not aa

    # Add 2 attachments
    assert aa.add(join(TEST_VAR_DIR, "apprise-test.gif"))
    assert aa.add(join(TEST_VAR_DIR, "apprise-test.png"))

    assert mock_request.call_count == 0
    assert ap_obj.notify(body="body", title="title", attach=aa)

    assert mock_request.call_count == 1

    # Our first item was truncated, so only 1 attachment
    details = mock_request.call_args_list[0]
    assert details[0][0] == "GET"
    dataset = json.loads(details[1]["data"])
    assert len(dataset["attachments"]) == 1

    # Reset our object
    mock_request.reset_mock()

    # our Apprise Object
    ap_obj = Apprise()

    # Add ourselves an object set to upstream
    ap_obj.add("json://localhost/?method=GET&overflow=upstream")

    # Create ourselves an attachment object
    aa = AppriseAttachment()

    # Add 2 attachments
    assert aa.add(join(TEST_VAR_DIR, "apprise-test.gif"))
    assert aa.add(join(TEST_VAR_DIR, "apprise-test.png"))

    assert mock_request.call_count == 0
    assert ap_obj.notify(body="body", title="title", attach=aa)

    assert mock_request.call_count == 1

    # Our item was not truncated, so all attachments
    details = mock_request.call_args_list[0]
    assert details[0][0] == "GET"
    dataset = json.loads(details[1]["data"])
    assert len(dataset["attachments"]) == 2


def test_apprise_attachment_instantiate():
    """
    API: AppriseAttachment.instantiate()

    """
    assert (
        AppriseAttachment.instantiate("file://?", suppress_exceptions=True)
        is None
    )

    assert (
        AppriseAttachment.instantiate("invalid://?", suppress_exceptions=True)
        is None
    )

    class BadAttachType(AttachBase):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

            # We fail whenever we're initialized
            raise TypeError()

    # Store our bad attachment type in our schema map
    A_MGR["bad"] = BadAttachType

    with pytest.raises(TypeError):
        AppriseAttachment.instantiate("bad://path", suppress_exceptions=False)

    # Same call but exceptions suppressed
    assert (
        AppriseAttachment.instantiate("bad://path", suppress_exceptions=True)
        is None
    )


def test_attachment_matrix_dynamic_importing(tmpdir):
    """
    API: Apprise() Attachment Matrix Importing

    """

    # Make our new path valid
    suite = tmpdir.mkdir("apprise_attach_test_suite")
    suite.join("__init__.py").write("")

    module_name = "badattach"

    # Create a base area to work within
    base = suite.mkdir(module_name)
    base.join("__init__.py").write("")

    # Test no app_id
    base.join("AttachBadFile1.py").write(
        cleandoc("""
        class AttachBadFile1:
            pass
        """)
    )

    # No class of the same name
    base.join("AttachBadFile2.py").write(
        cleandoc("""
        class BadClassName:
            pass
        """)
    )

    # Exception thrown
    base.join("AttachBadFile3.py").write("""raise ImportError()""")

    # Utilizes a schema:// already occupied (as string)
    base.join("AttachGoober.py").write(
        cleandoc("""
        from apprise import AttachBase
        class AttachGoober(AttachBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol
            protocol = 'http'

            # The default secure protocol
            secure_protocol = 'https'
        """)
    )

    # Utilizes a schema:// already occupied (as tuple)
    base.join("AttachBugger.py").write(
        cleandoc("""
        from apprise import AttachBase
        class AttachBugger(AttachBase):
            # This class tests the fact we have a new class name, but we're
            # trying to over-ride items previously used

            # The default simple (insecure) protocol
            protocol = ('http', 'bugger-test' )

            # The default secure protocol
            secure_protocol = ('https', 'bugger-tests')
        """)
    )

    A_MGR.load_modules(path=str(base), name=module_name)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from datetime import datetime
from json import dumps, loads

# Disable logging for a cleaner testing output
import logging
import os
from unittest import mock

from helpers import AppriseURLTester
import pytest
import requests

from apprise import Apprise, AppriseAttachment, NotifyType
from apprise.plugins.office365 import NotifyOffice365

logging.disable(logging.CRITICAL)

# Attachment Directory
TEST_VAR_DIR = os.path.join(os.path.dirname(__file__), "var")

# Our Testing URLs
apprise_url_tests = (
    ##################################
    # NotifyOffice365
    ##################################
    (
        "o365://",
        {
            # Missing tenant, client_id, secret, and targets!
            "instance": TypeError,
        },
    ),
    (
        "o365://:@/",
        {
            # invalid url
            "instance": TypeError,
        },
    ),
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            # invalid tenant
            tenant=",",
            cid="ab-cd-ef-gh",
            aid="user@example.com",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # Expected failure
            "instance": TypeError,
        },
    ),
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            # invalid client id
            cid="ab.",
            aid="user2@example.com",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # Expected failure
            "instance": TypeError,
        },
    ),
    (
        "o365://{tenant}/{cid}/{secret}/{targets}".format(
            # email not required if mode is set to self
            tenant="tenant",
            cid="ab-cd-ef-gh",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
            },
        },
    ),
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.edu",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                # For 'From:' Lookup
                "mail": "user@example.ca",
            },
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "azure://user@example.edu/t...t/a...h/****/email1@test.ca/"
            ),
            "force_debug": True,
        },
    ),
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            # Source can also be Object ID
            aid="hg-fe-dc-ba",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
                "displayName": "John",
            },
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "azure://hg-fe-dc-ba/t...t/a...h/****/email1@test.ca/"
            ),
        },
    ),
    # ObjectID Specified, but no targets
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            # Source can also be Object ID
            aid="hg-fe-dc-ba",
            secret="abcd/123/3343/@jack/test",
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
            },
            # No emails detected
            "notify_response": False,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "azure://hg-fe-dc-ba/t...t/a...h/****",
        },
    ),
    # ObjectID Specified, but no targets
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            # Source can also be Object ID
            aid="hg-fe-dc-ba",
            secret="abcd/123/3343/@jack/test",
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "userPrincipalName": "user@example.ca",
            },
            # No emails detected
            "notify_response": False,
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": "azure://hg-fe-dc-ba/t...t/a...h/****",
        },
    ),
    # test our arguments
    (
        "o365://_/?oauth_id={cid}&oauth_secret={secret}&tenant={tenant}"
        "&to={targets}&from={aid}".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.ca",
            secret="abcd/123/3343/@jack/test",
            targets="email1@test.ca",
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # Test what happens if a batch send fails to return a messageCount
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
            },
            # Our expected url(privacy=True) startswith() response:
            "privacy_url": (
                "azure://user@example.ca/t...t/a...h/****/email1@test.ca/"
            ),
        },
    ),
    # Test invalid JSON (no tenant defaults to email domain)
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.com",
            secret="abcd/123/3343/@jack/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # invalid JSON response
            "requests_response_text": "{",
            "notify_response": False,
        },
    ),
    # No Targets specified
    (
        "o365://{aid}/{tenant}/{cid}/{secret}".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.com",
            secret="abcd/123/3343/@jack/test",
        ),
        {
            # We're valid and good to go
            "instance": NotifyOffice365,
            # There were no targets to notify; so we use our own email
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "userPrincipalName": "user@example.ca",
            },
        },
    ),
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            cid="zz-zz-zz-zz",
            aid="user@example.com",
            secret="abcd/abc/dcba/@john/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            "instance": NotifyOffice365,
            # throw a bizarre code forcing us to fail to look it up
            "response": False,
            "requests_response_code": 999,
        },
    ),
    (
        "o365://{tenant}:{aid}/{cid}/{secret}/{targets}".format(
            tenant="tenant",
            cid="01-12-23-34",
            aid="user@example.com",
            secret="abcd/321/4321/@test/test",
            targets="/".join(["email1@test.ca"]),
        ),
        {
            "instance": NotifyOffice365,
            # Throws a series of i/o exceptions with this flag
            # is set and tests that we gracefully handle them
            "test_requests_exceptions": True,
        },
    ),
    # reply_to support — plain address
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}"
        "?reply_to=reply@example.org".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.edu",
            secret="abcd/123/3343/@jack/test",
            targets="email1@test.ca",
        ),
        {
            "instance": NotifyOffice365,
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
            },
            "privacy_url": (
                "azure://user@example.edu/t...t/a...h/****/email1@test.ca/"
            ),
        },
    ),
    # reply_to with URL-encoded (+) spaces in name
    (
        "o365://{aid}/{tenant}/{cid}/{secret}/{targets}"
        "?reply_to=Reply+Person+reply@example.org".format(
            tenant="tenant",
            cid="ab-cd-ef-gh",
            aid="user@example.edu",
            secret="abcd/123/3343/@jack/test",
            targets="email1@test.ca",
        ),
        {
            "instance": NotifyOffice365,
            "requests_response_text": {
                "expires_in": 2000,
                "access_token": "abcd1234",
                "mail": "user@example.ca",
            },
            "privacy_url": (
                "azure://user@example.edu/t...t/a...h/****/email1@test.ca/"
            ),
        },
    ),
)


def test_plugin_office365_urls():
    """NotifyOffice365() Apprise URLs."""

    # Run our general tests
    AppriseURLTester(tests=apprise_url_tests).run_all()


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_general(mock_get, mock_post):
    """NotifyOffice365() General Testing."""

    # Initialize some generic (but valid) tokens
    email = "user@example.net"
    tenant = "ff-gg-hh-ii-jj"
    client_id = "aa-bb-cc-dd-ee"
    secret = "abcd/1234/abcd@ajd@/test"
    targets = "target@example.com"

    # Prepare Mock return object
    payload = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For 'From:' Lookup
        "mail": "abc@example.ca",
        # For our Draft Email ID:
        "id": "draft-id-no",
    }
    response = mock.Mock()
    response.content = dumps(payload)
    response.status_code = requests.codes.ok
    mock_post.return_value = response
    mock_get.return_value = response

    # Instantiate our object
    obj = Apprise.instantiate(f"o365://{email}/{tenant}/{secret}/{targets}")

    assert isinstance(obj, NotifyOffice365)

    # Test our URL generation
    assert isinstance(obj.url(), str)

    # Test our notification
    assert obj.notify(title="title", body="test") is True

    # Instantiate our object
    obj = Apprise.instantiate(
        "o365://{email}/{tenant}/{client_id}/{secret}/{targets}"
        "?bcc={bcc}&cc={cc}".format(
            tenant=tenant,
            email=email,
            client_id=client_id,
            secret=secret,
            targets=targets,
            # Test the cc and bcc list (use good and bad email)
            cc="Chuck Norris cnorris@yahoo.ca, Sauron@lotr.me, invalid@!",
            bcc="Bruce Willis bwillis@hotmail.com, Frodo@lotr.me invalid@!",
        )
    )

    assert isinstance(obj, NotifyOffice365)

    # Test our URL generation
    assert isinstance(obj.url(), str)

    # Test our notification
    assert obj.notify(title="title", body="test") is True

    with pytest.raises(TypeError):
        # No secret
        NotifyOffice365(
            email=email,
            client_id=client_id,
            tenant=tenant,
            secret=None,
            targets=None,
        )

    # One of the targets are invalid
    obj = NotifyOffice365(
        email=email,
        client_id=client_id,
        tenant=tenant,
        secret=secret,
        targets=("Management abc@gmail.com", "garbage"),
    )
    # Test our notification (this will work and only notify abc@gmail.com)
    assert obj.notify(title="title", body="test") is True

    # all of the targets are invalid
    obj = NotifyOffice365(
        email=email,
        client_id=client_id,
        tenant=tenant,
        secret=secret,
        targets=("invalid", "garbage"),
    )

    # Test our notification (which will fail because of no entries)
    assert obj.notify(title="title", body="test") is False


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_authentication(mock_get, mock_post):
    """NotifyOffice365() Authentication Testing."""

    # Initialize some generic (but valid) tokens
    tenant = "ff-gg-hh-ii-jj"
    email = "user@example.net"
    client_id = "aa-bb-cc-dd-ee"
    secret = "abcd/1234/abcd@ajd@/test"
    targets = "target@example.com"

    # Prepare Mock return object
    authentication_okay = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
    }
    authentication_failure = {
        "error": "invalid_scope",
        "error_description": "AADSTS70011: Blah... Blah Blah... Blah",
        "error_codes": [70011],
        "timestamp": "2020-01-09 02:02:12Z",
        "trace_id": "255d1aef-8c98-452f-ac51-23d051240864",
        "correlation_id": "fb3d2015-bc17-4bb9-bb85-30c5cf1aaaa7",
    }
    response = mock.Mock()
    response.content = dumps(authentication_okay)
    response.status_code = requests.codes.ok
    mock_post.return_value = response
    mock_get.return_value = response

    # Instantiate our object
    obj = Apprise.instantiate(
        f"azure://{email}/{tenant}/{client_id}/{secret}/{targets}"
    )

    assert isinstance(obj, NotifyOffice365)

    # Authenticate
    assert obj.authenticate() is True

    # We're already authenticated
    assert obj.authenticate() is True

    # Expire our token
    obj.token_expiry = datetime.now()

    # Re-authentiate
    assert obj.authenticate() is True

    # Change our response
    response.status_code = 400

    # We'll fail to send a notification now...
    assert obj.notify(title="title", body="test") is False

    # Expire our token
    obj.token_expiry = datetime.now()

    # Set a failure response
    response.content = dumps(authentication_failure)

    # We will fail to authenticate at this point
    assert obj.authenticate() is False

    # Notifications will also fail in this case
    assert obj.notify(title="title", body="test") is False

    # We will fail to authenticate with invalid data

    invalid_auth_entries = authentication_okay.copy()
    invalid_auth_entries["expires_in"] = "garbage"
    response.content = dumps(invalid_auth_entries)
    response.status_code = requests.codes.ok
    assert obj.authenticate() is False

    invalid_auth_entries["expires_in"] = None
    response.content = dumps(invalid_auth_entries)
    assert obj.authenticate() is False

    invalid_auth_entries["expires_in"] = ""
    response.content = dumps(invalid_auth_entries)
    assert obj.authenticate() is False

    del invalid_auth_entries["expires_in"]
    response.content = dumps(invalid_auth_entries)
    assert obj.authenticate() is False


@mock.patch("requests.put")
@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_queries(mock_post, mock_get, mock_put):
    """NotifyOffice365() General Queries."""

    # Initialize some generic (but valid) tokens
    source = "abc-1234-object-id"
    tenant = "ff-gg-hh-ii-jj"
    client_id = "aa-bb-cc-dd-ee"
    secret = "abcd/1234/abcd@ajd@/test"
    targets = "target@example.ca"

    # Prepare Mock return object
    payload = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For 'From:' Lookup (email)
        "mail": "user@example.edu",
        # For 'From:' Lookup (name)
        "displayName": "John",
        # For our Draft Email ID:
        "id": "draft-id-no",
        # For FIle Uploads
        "uploadUrl": "https://my.url.path/",
    }

    okay_response = mock.Mock()
    okay_response.content = dumps(payload)
    okay_response.status_code = requests.codes.ok
    mock_post.return_value = okay_response
    mock_put.return_value = okay_response

    bad_response = mock.Mock()
    bad_response.content = dumps(payload)
    bad_response.status_code = requests.codes.forbidden

    # Assign our GET a bad response so we fail to look up the user
    mock_get.return_value = bad_response

    #
    # Early exit: source is an Object ID (GUID) and lookup fails
    #
    guid_source = "e2a9db46-5714-44c2-87bf-c33bf236f42a"

    not_found_payload = {
        "error": {
            "code": "Request_ResourceNotFound",
            "message": "Resource not found.",
        }
    }
    not_found_response = mock.Mock()
    not_found_response.content = dumps(not_found_payload)
    not_found_response.status_code = requests.codes.not_found

    # Force the sender lookup to fail for a GUID based source
    mock_get.return_value = not_found_response

    obj = Apprise.instantiate(
        f"azure://{guid_source}/{tenant}/{client_id}{secret}/{targets}"
    )
    assert isinstance(obj, NotifyOffice365)

    # Notify must exit early since the source Object ID could not be resolved
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is False
    )

    # 1) Authenticate (POST token)
    assert mock_post.call_count == 1
    # 2) Attempt user lookup (GET)
    assert mock_get.call_count == 1

    mock_post.reset_mock()
    mock_get.reset_mock()

    # Restore GET failure behaviour used by remaining tests
    mock_get.return_value = bad_response

    # Instantiate our object
    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    assert isinstance(obj, NotifyOffice365)

    # We can still send a notification even if we can't look up the email
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )

    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://graph.microsoft.com/v1.0/users/abc-1234-object-id/sendMail"
    )
    payload = loads(mock_post.call_args_list[1][1]["data"])
    assert payload == {
        "message": {
            "subject": "title",
            "body": {
                "contentType": "HTML",
                "content": "body",
            },
            "toRecipients": [
                {"emailAddress": {"address": "target@example.ca"}}
            ],
        },
        "saveToSentItems": "true",
    }
    mock_post.reset_mock()

    # Now test a case where we just couldn't get any email details from the
    # payload returned

    # Prepare Mock return object
    temp_payload = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For our Draft Email ID:
        "id": "draft-id-no",
        # For FIle Uploads
        "uploadUrl": "https://my.url.path/",
    }

    bad_response.content = dumps(temp_payload)
    bad_response.status_code = requests.codes.okay
    mock_get.return_value = bad_response

    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    assert isinstance(obj, NotifyOffice365)

    # We can still send a notification even if we can't look up the email
    assert (
        obj.notify(body="body", title="title", notify_type=NotifyType.INFO)
        is True
    )


@mock.patch("requests.put")
@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_attachments(mock_post, mock_get, mock_put):
    """NotifyOffice365() Attachments."""

    # Initialize some generic (but valid) tokens
    source = "user@example.net"
    tenant = "ff-gg-hh-ii-jj"
    client_id = "aa-bb-cc-dd-ee"
    secret = "abcd/1234/abcd@ajd@/test"
    targets = "target@example.com"

    # Prepare Mock return object
    payload = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For 'From:' Lookup
        "mail": "user@example.edu",
        # For our Draft Email ID:
        "id": "draft-id-no",
        # For FIle Uploads
        "uploadUrl": "https://my.url.path/",
    }
    okay_response = mock.Mock()
    okay_response.content = dumps(payload)
    okay_response.status_code = requests.codes.ok
    mock_post.return_value = okay_response
    mock_get.return_value = okay_response
    mock_put.return_value = okay_response

    # Instantiate our object
    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    assert isinstance(obj, NotifyOffice365)

    # Test Valid Attachment
    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[0][1]["headers"].get("Content-Type")
        == "application/x-www-form-urlencoded"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/sendMail"
    )
    assert (
        mock_post.call_args_list[1][1]["headers"].get("Content-Type")
        == "application/json"
    )
    mock_post.reset_mock()

    # Test Authentication Failure
    obj = Apprise.instantiate(
        "azure://{source}/{tenant}/{client_id}{secret}/{targets}".format(
            client_id=client_id,
            tenant=tenant,
            source="object-id-requiring-lookup",
            secret=secret,
            targets=targets,
        )
    )

    bad_response = mock.Mock()
    bad_response.content = dumps(payload)
    bad_response.status_code = requests.codes.forbidden
    mock_post.return_value = bad_response

    assert isinstance(obj, NotifyOffice365)
    # Authentication will fail
    assert (
        obj.notify(
            body="auth-fail", title="title", notify_type=NotifyType.INFO
        )
        is False
    )
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/ff-gg-hh-ii-jj/oauth2/v2.0/token"
    )
    mock_post.reset_mock()

    #
    # Test invalid attachment
    #

    # Instantiate our object
    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    assert isinstance(obj, NotifyOffice365)

    mock_post.return_value = okay_response
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=path,
        )
        is False
    )
    assert mock_post.call_count == 0
    mock_post.reset_mock()

    with mock.patch("base64.b64encode", side_effect=OSError()):
        # We can't send the message if we fail to parse the data
        assert (
            obj.notify(
                body="body",
                title="title",
                notify_type=NotifyType.INFO,
                attach=attach,
            )
            is False
        )
    assert mock_post.call_count == 0
    mock_post.reset_mock()

    #
    # Test case where we can't authenticate
    #
    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    # Force a smaller attachment size forcing us to create an attachment
    obj.outlook_attachment_inline_max = 50

    assert isinstance(obj, NotifyOffice365)

    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)
    mock_post.return_value = bad_response
    assert obj.upload_attachment(attach[0], "id") is False
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/ff-gg-hh-ii-jj/oauth2/v2.0/token"
    )

    mock_post.reset_mock()

    mock_post.side_effect = (okay_response, bad_response)
    mock_post.return_value = None
    assert obj.upload_attachment(attach[0], "id") is False
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/ff-gg-hh-ii-jj/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/"
        + "message/id/attachments/createUploadSession"
    )

    mock_post.reset_mock()
    # Return our status
    mock_post.side_effect = None

    # Prepare Mock return object
    payload_no_upload_url = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For 'From:' Lookup
        "mail": "user@example.edu",
        # For our Draft Email ID:
        "id": "draft-id-no",
    }
    tmp_response = mock.Mock()
    tmp_response.content = dumps(payload_no_upload_url)
    tmp_response.status_code = requests.codes.ok
    mock_post.return_value = tmp_response

    assert obj.upload_attachment(attach[0], "id") is False
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/"
        + "message/id/attachments/createUploadSession"
    )

    mock_post.reset_mock()
    # Return our status
    mock_post.side_effect = None
    mock_post.return_value = okay_response

    obj = Apprise.instantiate(
        f"azure://{source}/{tenant}/{client_id}{secret}/{targets}"
    )

    # Force a smaller attachment size forcing us to create an attachment
    obj.outlook_attachment_inline_max = 50

    assert isinstance(obj, NotifyOffice365)

    # We now have to prepare sepparate session attachments using draft emails
    assert (
        obj.notify(
            body="body",
            title="title-test",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # Large Attachments
    assert mock_post.call_count == 4
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/ff-gg-hh-ii-jj/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/messages"
    )
    assert (
        mock_post.call_args_list[2][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/"
        + "message/draft-id-no/attachments/createUploadSession"
    )
    assert (
        mock_post.call_args_list[3][0][0]
        == "https://graph.microsoft.com/v1.0/users/"
        + f"{source}/messages/draft-id-no/send"
    )
    mock_post.reset_mock()

    #
    # Handle another case where can't upload the attachment at all
    #
    path = os.path.join(TEST_VAR_DIR, "/invalid/path/to/an/invalid/file.jpg")
    bad_attach = AppriseAttachment(path)
    assert obj.upload_attachment(bad_attach[0], "id") is False

    mock_post.reset_mock()
    #
    # Handle test case where we can't send the draft email after everything
    # has been prepared
    #
    mock_post.return_value = None
    mock_post.side_effect = (okay_response, okay_response, bad_response)
    assert (
        obj.notify(
            body="body",
            title="title-test",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is False
    )

    assert mock_post.call_count == 3
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/messages"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/"
        + "message/draft-id-no/attachments/createUploadSession"
    )
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://graph.microsoft.com/v1.0/users/"
        + f"{source}/messages/draft-id-no/send"
    )
    mock_post.reset_mock()
    mock_post.side_effect = None
    mock_post.return_value = okay_response

    #
    # Handle test case where we can not upload chunks
    #
    mock_put.return_value = bad_response

    # We now have to prepare sepparate session attachments using draft emails
    assert (
        obj.notify(
            body="body",
            title="title-no-chunk",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is False
    )

    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/messages"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/"
        + "message/draft-id-no/attachments/createUploadSession"
    )

    mock_put.return_value = okay_response
    mock_post.reset_mock()

    # Prepare Mock return object
    payload_missing_id = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        # For 'From:' Lookup
        "mail": "user@example.edu",
        # For FIle Uploads
        "uploadUrl": "https://my.url.path/",
    }
    temp_response = mock.Mock()
    temp_response.content = dumps(payload_missing_id)
    temp_response.status_code = requests.codes.ok
    mock_post.return_value = temp_response

    # We could not acquire an attachment id, so we'll fail to send our
    # notification
    assert (
        obj.notify(
            body="body",
            title="title-test",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is False
    )

    # Large Attachments
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://graph.microsoft.com/v1.0/users/user@example.net/messages"
    )

    mock_post.reset_mock()

    # Reset attachment size
    obj.outlook_attachment_inline_max = 50 * 1024 * 1024
    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # already authenticated
    assert mock_post.call_count == 1
    assert (
        mock_post.call_args_list[0][0][0]
        == f"https://graph.microsoft.com/v1.0/users/{source}/sendMail"
    )
    mock_post.reset_mock()


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_reply_to(mock_post, mock_get):
    """NotifyOffice365() reply_to parameter support."""

    tenant = "ff-gg-hh-ii-jj"
    email = "user@example.net"
    client_id = "aa-bb-cc-dd-ee"
    secret = "abcd/1234/abcd@ajd@/test"
    targets = "target@example.com"

    payload = {
        "token_type": "Bearer",
        "expires_in": 6000,
        "access_token": "abcd1234",
        "mail": "user@example.net",
    }
    okay_response = mock.Mock()
    okay_response.content = dumps(payload)
    okay_response.status_code = requests.codes.ok
    mock_post.return_value = okay_response
    mock_get.return_value = okay_response

    # named reply_to via direct instantiation
    obj = NotifyOffice365(
        source=email,
        tenant=tenant,
        client_id=client_id,
        secret=secret,
        targets=[targets],
        reply_to=["Reply Person <reply@example.org>"],
    )

    assert isinstance(obj, NotifyOffice365)
    assert len(obj.reply_to) == 1
    assert "reply@example.org" in obj.reply_to

    # url() round-trip includes reply_to
    url = obj.url()
    assert "reply_to=" in url

    # send() adds replyTo with name to the payload
    assert obj.notify(title="title", body="test") is True
    assert mock_post.call_count == 2
    sent_payload = loads(mock_post.call_args_list[1][1]["data"])
    assert "replyTo" in sent_payload["message"]
    assert (
        sent_payload["message"]["replyTo"][0]["emailAddress"]["address"]
        == "reply@example.org"
    )
    assert (
        sent_payload["message"]["replyTo"][0]["emailAddress"]["name"]
        == "Reply Person"
    )
    mock_post.reset_mock()

    # reply_to without a name via URL query string
    obj2 = Apprise.instantiate(
        "o365://{email}/{tenant}/{client_id}/{secret}/{targets}"
        "?reply_to=reply@example.org".format(
            tenant=tenant,
            email=email,
            client_id=client_id,
            secret=secret,
            targets=targets,
        )
    )

    assert isinstance(obj2, NotifyOffice365)
    assert obj2.notify(title="t", body="b") is True
    sent_payload = loads(mock_post.call_args_list[1][1]["data"])
    assert "replyTo" in sent_payload["message"]
    assert "name" not in sent_payload["message"]["replyTo"][0]["emailAddress"]
    mock_post.reset_mock()

    # invalid reply_to address is dropped silently
    obj3 = NotifyOffice365(
        source=email,
        tenant=tenant,
        client_id=client_id,
        secret=secret,
        targets=[targets],
        reply_to=["not-a-valid-email"],
    )
    assert isinstance(obj3, NotifyOffice365)
    assert len(obj3.reply_to) == 0
    # No replyTo in payload when reply_to is empty
    assert obj3.notify(title="t", body="b") is True
    sent_payload = loads(mock_post.call_args_list[1][1]["data"])
    assert "replyTo" not in sent_payload["message"]
    mock_post.reset_mock()

    # parse_url picks up ?reply_to=
    parsed = NotifyOffice365.parse_url(
        "o365://{email}/{tenant}/{client_id}/{secret}/{targets}"
        "?reply_to=reply@example.org".format(
            tenant=tenant,
            email=email,
            client_id=client_id,
            secret=secret,
            targets=targets,
        )
    )
    assert parsed is not None
    assert "reply_to" in parsed
    assert "reply@example.org" in parsed["reply_to"]


@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_personal_mode(mock_post, mock_get):
    """NotifyOffice365() personal mode (refresh_token / consumer accounts)."""
    from apprise.plugins.office365 import (
        Office365Mode,
    )

    client_id = "aa-bb-cc-dd-ee"
    seed_token = "seed-refresh-token-abc123"
    new_token = "rotated-refresh-token-xyz789"
    access_token = "access-token-live"

    auth_payload = {
        "token_type": "Bearer",
        "expires_in": 3600,
        "access_token": access_token,
        "refresh_token": new_token,
    }
    okay_response = mock.Mock()
    okay_response.content = dumps(auth_payload)
    okay_response.status_code = requests.codes.ok
    bad_response = mock.Mock()
    bad_response.content = dumps({"error": "bad"})
    bad_response.status_code = requests.codes.internal_server_error

    mock_post.return_value = okay_response
    mock_get.return_value = okay_response

    #
    # Auto-detection: @live.com triggers personal mode
    #
    obj = NotifyOffice365(
        source="user@live.com",
        client_id=client_id,
        secret=seed_token,
    )
    assert isinstance(obj, NotifyOffice365)
    assert obj.mode == Office365Mode.PERSONAL
    assert obj.tenant is None
    assert obj.from_email == "user@live.com"
    assert obj.source_is_object_id is False

    # All known personal domains trigger personal mode
    for domain in ("hotmail.com", "outlook.com", "live.com", "msn.com"):
        o = NotifyOffice365(
            source=f"user@{domain}",
            client_id=client_id,
            secret=seed_token,
        )
        assert o.mode == Office365Mode.PERSONAL, (
            f"expected personal mode for {domain}"
        )

    #
    # Explicit mode=personal on a non-personal domain
    #
    obj_explicit = NotifyOffice365(
        source="user@example.com",
        client_id=client_id,
        secret=seed_token,
        mode=Office365Mode.PERSONAL,
    )
    assert obj_explicit.mode == Office365Mode.PERSONAL

    #
    # Explicit mode=org on a personal domain — still requires valid email
    # source (no tenant needed for the personal flow, but org requires tenant)
    #
    with pytest.raises(TypeError):
        # tenant is None but mode=org demands a valid tenant
        NotifyOffice365(
            source="user@live.com",
            client_id=client_id,
            secret=seed_token,
            mode=Office365Mode.ORG,
        )

    #
    # Personal mode requires source to be a valid email
    #
    with pytest.raises(TypeError):
        NotifyOffice365(
            source="not-an-email",
            client_id=client_id,
            secret=seed_token,
            mode=Office365Mode.PERSONAL,
        )

    #
    # Invalid mode string raises TypeError
    #
    with pytest.raises(TypeError):
        NotifyOffice365(
            source="user@live.com",
            client_id=client_id,
            secret=seed_token,
            mode="invalid-mode",
        )

    #
    # send() uses me/sendMail and me/messages (not users/{source}/...)
    #
    obj = NotifyOffice365(
        source="user@live.com",
        client_id=client_id,
        secret=seed_token,
        targets=["target@example.com"],
    )
    assert obj.mode == Office365Mode.PERSONAL
    assert obj.notify(title="title", body="body") is True

    # POST 0: token refresh  POST 1: sendMail
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/consumers/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://graph.microsoft.com/v1.0/me/sendMail"
    )
    mock_post.reset_mock()

    #
    # Refresh token rotation: store holds the new token; url() emits it
    #
    stored = obj.store.get("refresh_token")
    assert stored == new_token

    generated_url = obj.url()
    assert new_token in generated_url
    assert seed_token not in generated_url
    assert "mode=personal" in generated_url

    # Privacy URL masks the token
    private_url = obj.url(privacy=True)
    assert new_token not in private_url
    assert seed_token not in private_url

    #
    # Auth failure (bad token response) propagates correctly
    #
    mock_post.return_value = bad_response
    obj2 = NotifyOffice365(
        source="user@hotmail.com",
        client_id=client_id,
        secret=seed_token,
    )
    assert obj2.notify(title="t", body="b") is False
    mock_post.reset_mock()
    mock_post.return_value = okay_response

    #
    # Auth response missing expires_in
    #
    broken_payload = {"access_token": access_token}
    broken_response = mock.Mock()
    broken_response.content = dumps(broken_payload)
    broken_response.status_code = requests.codes.ok
    mock_post.return_value = broken_response
    obj3 = NotifyOffice365(
        source="user@live.com",
        client_id=client_id,
        secret=seed_token,
    )
    assert obj3.notify(title="t", body="b") is False
    mock_post.reset_mock()
    mock_post.return_value = okay_response

    #
    # parse_url: personal domain auto-detects without mode= param
    #
    parsed = NotifyOffice365.parse_url(
        f"o365://user@live.com/{client_id}/{seed_token}/"
    )
    assert parsed is not None
    assert parsed["mode"] == Office365Mode.PERSONAL
    assert parsed["tenant"] is None
    assert parsed["client_id"] == client_id
    assert parsed["secret"] == seed_token

    # Reconstructed object from parse_url
    obj4 = Apprise.instantiate(
        f"o365://user@live.com/{client_id}/{seed_token}/"
    )
    assert isinstance(obj4, NotifyOffice365)
    assert obj4.mode == Office365Mode.PERSONAL

    #
    # parse_url: explicit ?mode=personal on non-personal domain
    #
    parsed2 = NotifyOffice365.parse_url(
        f"o365://user@example.com/{client_id}/{seed_token}/?mode=personal"
    )
    assert parsed2 is not None
    assert parsed2["mode"] == Office365Mode.PERSONAL

    #
    # parse_url: explicit ?mode=org on personal domain
    #
    parsed3 = NotifyOffice365.parse_url(
        f"o365://user@live.com/tenant-id/{client_id}/{seed_token}/?mode=org"
    )
    assert parsed3 is not None
    assert parsed3["mode"] == Office365Mode.ORG
    assert parsed3["tenant"] == "tenant-id"

    #
    # url() round-trip for personal mode
    #
    obj5 = NotifyOffice365(
        source="user@outlook.com",
        client_id=client_id,
        secret=seed_token,
        targets=["recipient@example.com"],
    )
    url = obj5.url()
    assert "mode=personal" in url
    assert "outlook.com" in url
    assert client_id in url
    # No tenant segment
    assert "tenant" not in url

    # Round-trip: parse and reconstruct
    parsed4 = NotifyOffice365.parse_url(url)
    assert parsed4 is not None
    assert parsed4["mode"] == Office365Mode.PERSONAL
    obj6 = Apprise.instantiate(url)
    assert isinstance(obj6, NotifyOffice365)
    assert obj6.mode == Office365Mode.PERSONAL

    #
    # parse_url: unrecognised ?mode= value falls back to auto-detection
    # (covers the branch where _mode is NOT in OFFICE365_MODES)
    #
    parsed5 = NotifyOffice365.parse_url(
        f"o365://user@live.com/{client_id}/{seed_token}/?mode=bogus-value"
    )
    assert parsed5 is not None
    # bogus mode is ignored; auto-detection kicks in → personal
    assert parsed5["mode"] == Office365Mode.PERSONAL

    #
    # Personal mode supports cc/bcc/reply_to and emits them via url()
    #
    mock_post.reset_mock()
    mock_post.return_value = okay_response
    obj7 = NotifyOffice365(
        source="user@outlook.com",
        client_id=client_id,
        secret=seed_token,
        targets=["target@example.com"],
        cc=["Carbon Copy <cc@example.com>"],
        bcc=["Blind Copy <bcc@example.com>"],
        reply_to=["Reply Person <reply@example.org>"],
    )
    assert isinstance(obj7, NotifyOffice365)

    personal_url = obj7.url()
    assert "mode=personal" in personal_url
    assert "cc=Carbon%20Copy%3Acc%40example.com" in personal_url
    assert "bcc=Blind%20Copy%3Abcc%40example.com" in personal_url
    assert "reply_to=Reply%20Person%3Areply%40example.org" in personal_url

    assert obj7.notify(title="title", body="body") is True
    assert mock_post.call_count == 2
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://graph.microsoft.com/v1.0/me/sendMail"
    )
    sent_payload = loads(mock_post.call_args_list[1][1]["data"])
    assert sent_payload["message"]["ccRecipients"][0]["emailAddress"] == {
        "address": "cc@example.com",
        "name": "Carbon Copy",
    }
    assert sent_payload["message"]["bccRecipients"][0]["emailAddress"] == {
        "address": "bcc@example.com",
        "name": "Blind Copy",
    }
    assert sent_payload["message"]["replyTo"][0]["emailAddress"] == {
        "address": "reply@example.org",
        "name": "Reply Person",
    }

    #
    # Auth succeeds but server returns no new refresh_token
    # (covers the branch where `if new_refresh:` is falsy)
    #
    no_rotate_payload = {
        "token_type": "Bearer",
        "expires_in": 3600,
        "access_token": access_token,
        # no "refresh_token" key
    }
    no_rotate_response = mock.Mock()
    no_rotate_response.content = dumps(no_rotate_payload)
    no_rotate_response.status_code = requests.codes.ok
    send_response = mock.Mock()
    send_response.content = dumps({})
    send_response.status_code = requests.codes.accepted

    mock_post.side_effect = [no_rotate_response, send_response]
    obj_nr = NotifyOffice365(
        source="user@live.com",
        client_id=client_id,
        secret=seed_token,
        targets=["target@example.com"],
    )
    assert obj_nr.notify(title="t", body="b") is True
    # No new token stored — seed_token still in url()
    assert obj_nr.store.get("refresh_token") is None
    mock_post.reset_mock()
    mock_post.side_effect = None
    mock_post.return_value = okay_response


@mock.patch("requests.put")
@mock.patch("requests.get")
@mock.patch("requests.post")
def test_plugin_office365_personal_large_attachment(
    mock_post, mock_get, mock_put
):
    """Personal mode large-attachment path (me/messages draft + send)."""
    from apprise.plugins.office365 import Office365Mode

    client_id = "aa-bb-cc-dd-ee"
    seed_token = "seed-refresh-token-abc123"
    new_token = "rotated-refresh-token-xyz789"
    access_token = "access-token-live"

    # Response that covers auth tokens, draft ID, and upload URL
    payload = {
        "token_type": "Bearer",
        "expires_in": 3600,
        "access_token": access_token,
        "refresh_token": new_token,
        "id": "draft-id-no",
        "uploadUrl": "https://upload.url.example/session",
    }
    okay_response = mock.Mock()
    okay_response.content = dumps(payload)
    okay_response.status_code = requests.codes.ok
    mock_post.return_value = okay_response
    mock_get.return_value = okay_response
    mock_put.return_value = okay_response

    path = os.path.join(TEST_VAR_DIR, "apprise-test.gif")
    attach = AppriseAttachment(path)

    obj = NotifyOffice365(
        source="user@live.com",
        client_id=client_id,
        secret=seed_token,
        targets=["target@example.com"],
    )
    assert obj.mode == Office365Mode.PERSONAL

    # Force all attachments to be "large" so the draft path is used
    obj.outlook_attachment_inline_max = 50

    assert (
        obj.notify(
            body="body",
            title="title",
            notify_type=NotifyType.INFO,
            attach=attach,
        )
        is True
    )

    # POST 0: token refresh  POST 1: draft create  POST 2: upload session
    # POST 3: send draft
    assert mock_post.call_count == 4
    assert (
        mock_post.call_args_list[0][0][0]
        == "https://login.microsoftonline.com/consumers/oauth2/v2.0/token"
    )
    assert (
        mock_post.call_args_list[1][0][0]
        == "https://graph.microsoft.com/v1.0/me/messages"
    )
    assert (
        mock_post.call_args_list[2][0][0]
        == "https://graph.microsoft.com/v1.0/me/"
        "message/draft-id-no/attachments/createUploadSession"
    )
    assert (
        mock_post.call_args_list[3][0][0] == "https://graph.microsoft.com/v1.0"
        "/me/messages/draft-id-no/send"
    )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2026, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import logging
import ssl
from unittest import mock

import pytest

from apprise.exception import AppriseException, AppriseInvalidData
from apprise.utils.socket import AppriseSocketError, SocketTransport

# Disable logging for a cleaner testing output
logging.disable(logging.CRITICAL)


class _DummyFile:
    def __init__(self) -> None:
        self.flushed = 0
        self.closed = 0

    def flush(self) -> None:
        self.flushed += 1

    def close(self) -> None:
        self.closed += 1


class _DummySocket:
    """
    Minimal fake socket object, supports the subset used by SocketTransport.
    """

    def __init__(self) -> None:
        self.blocking = True
        self.timeout = None
        self.closed = False
        self.shutdown_called = 0

        self._recv_side_effect = None
        self._send_side_effect = None

    def setsockopt(self, *_args, **_kwargs) -> None:
        return None

    def bind(self, *_args, **_kwargs) -> None:
        return None

    def fileno(self) -> int:
        return 1

    def settimeout(self, value):
        self.timeout = value

    def setblocking(self, value):
        self.blocking = bool(value)

    def connect(self, *_args, **_kwargs) -> None:
        return None

    def shutdown(self, *_args, **_kwargs) -> None:
        self.shutdown_called += 1

    def close(self) -> None:
        self.closed = True

    def getsockname(self):
        return ("127.0.0.1", 12345)

    def getpeername(self):
        return ("203.0.113.10", 6667)

    def makefile(self, *_args, **_kwargs):
        return _DummyFile()

    def recv(self, *_args, **_kwargs):
        if self._recv_side_effect is not None:
            raise self._recv_side_effect
        return b"data"

    def send(self, *_args, **_kwargs):
        if self._send_side_effect is not None:
            raise self._send_side_effect
        return 4


def test_utils_socket_timeout_coerce():
    """SocketTransport() Timeout Coercion."""
    # None => None, None
    s = SocketTransport("example.com", 1, timeout=None)
    assert s._connect_timeout is None
    assert s._read_timeout is None

    # float => both
    s = SocketTransport("example.com", 1, timeout=2.5)
    assert s._connect_timeout == pytest.approx(2.5)
    assert s._read_timeout == pytest.approx(2.5)

    # tuple => (connect, read)
    s = SocketTransport("example.com", 1, timeout=(1.0, 3.0))
    assert s._connect_timeout == pytest.approx(1.0)
    assert s._read_timeout == pytest.approx(3.0)

    # tuple with None is allowed
    s = SocketTransport("example.com", 1, timeout=(None, 3.0))
    assert s._connect_timeout is None
    assert s._read_timeout == pytest.approx(3.0)

    # invalid types
    with pytest.raises(AppriseInvalidData):
        SocketTransport("example.com", 1, timeout="bad")

    with pytest.raises(AppriseInvalidData):
        SocketTransport("example.com", 1, timeout=(1.0, 2.0, 3.0))

    with pytest.raises(AppriseInvalidData):
        SocketTransport("example.com", 1, timeout=-1.0)


def test_utils_socket_properties_and_close_paths():
    """SocketTransport() Properties And Close Paths."""
    s = SocketTransport("example.com", 1)
    assert s.connected is False
    assert s.is_tls is False

    # Seed with dummy socket + wrappers
    s._sock = _DummySocket()
    s._rfile = _DummyFile()
    s._wfile = _DummyFile()
    s._is_tls = True
    s.local_addr = ("1.1.1.1", 1)
    s.remote_addr = ("2.2.2.2", 2)

    s.close()
    assert s._sock is None
    assert s._rfile is None
    assert s._wfile is None
    assert s._is_tls is False
    assert s.local_addr is None
    assert s.remote_addr is None


def test_utils_socket_read_write_no_socket():
    """SocketTransport() Can Read/Write - No Socket."""
    s = SocketTransport("example.com", 1)
    assert s.can_read() is None
    assert s.can_write() is None


def test_utils_socket_can_read_write_select_error():
    """SocketTransport() Can Read/Write - select() Error."""
    s = SocketTransport("example.com", 1)
    s._sock = _DummySocket()

    with mock.patch("select.select", side_effect=OSError()):
        assert s.can_read() is None
        assert s._sock is None

    s._sock = _DummySocket()
    with mock.patch("select.select", side_effect=OSError()):
        assert s.can_write() is None
        assert s._sock is None


def test_utils_socket_can_read_write_socket_close():
    """SocketTransport() Can Read/Write - Socket Close."""
    s = SocketTransport("example.com", 1)
    s._sock = _DummySocket()

    # Exceptional socket list returned triggers close()
    with mock.patch("select.select", return_value=([], [], [s._sock])):
        assert s.can_read() is None
        assert s._sock is None

    s._sock = _DummySocket()
    with mock.patch("select.select", return_value=([], [], [s._sock])):
        assert s.can_write() is None
        assert s._sock is None


def test_utils_socket_server_hostname_for_tls():
    """SocketTransport() Server Hostname For TLS."""
    s = SocketTransport("irc.example.com", 1, verify=True)
    assert s._server_hostname_for_tls() == "irc.example.com"

    s = SocketTransport("irc.example.com", 1, verify=False)
    assert s._server_hostname_for_tls() == "irc.example.com"


def test_utils_socket_server_hostname_for_tls_ip_reverse():
    """SocketTransport() Server Hostname For TLS IP Reverse DNS"""
    s = SocketTransport("203.0.113.10", 1, verify=True)

    with mock.patch(
        "socket.gethostbyaddr", return_value=("irc.example.com.", [], [])
    ):
        assert s._server_hostname_for_tls() == "irc.example.com"

    with mock.patch("socket.gethostbyaddr", side_effect=OSError()):
        assert s._server_hostname_for_tls() == "203.0.113.10"

    # verify=False => no lookup
    s = SocketTransport("203.0.113.10", 1, verify=False)
    with mock.patch("socket.gethostbyaddr") as m:
        assert s._server_hostname_for_tls() == "203.0.113.10"
        m.assert_not_called()


def test_utils_socket_build_ssl_context():
    """SocketTransport() Build SSL Context."""

    class _DummySSLContext:
        def __init__(self, *, raise_on_minimum: bool) -> None:
            self.raise_on_minimum = raise_on_minimum
            self.options = 0
            self.check_hostname = None
            self.verify_mode = None
            self._minimum_version = None

        @property
        def minimum_version(self):
            return self._minimum_version

        @minimum_version.setter
        def minimum_version(self, value):
            if self.raise_on_minimum:
                raise RuntimeError("minimum_version not supported")
            self._minimum_version = value

    # verify=True path with minimum_version support
    s = SocketTransport("example.com", 1, verify=True)
    ctx = _DummySSLContext(raise_on_minimum=False)

    with (
        mock.patch("certifi.where", return_value="/tmp/ca.pem"),
        mock.patch("ssl.create_default_context", return_value=ctx) as m,
    ):
        result = s._build_ssl_context()
        assert result is ctx
        m.assert_called_once()
        assert ctx.check_hostname is True
        assert ctx.verify_mode == ssl.CERT_REQUIRED

        # TLS 1.2+ preferred path
        assert ctx.minimum_version == ssl.TLSVersion.TLSv1_2

        # Compression disabled when supported
        if hasattr(ssl, "OP_NO_COMPRESSION"):
            assert (ctx.options & ssl.OP_NO_COMPRESSION) != 0

    # verify=False path still enforces TLS 1.2+
    s = SocketTransport("example.com", 1, verify=False)
    ctx = _DummySSLContext(raise_on_minimum=False)

    with mock.patch("ssl.create_default_context", return_value=ctx):
        result = s._build_ssl_context()
        assert result is ctx
        assert ctx.check_hostname is False
        assert ctx.verify_mode == ssl.CERT_NONE
        assert ctx.minimum_version == ssl.TLSVersion.TLSv1_2

    # Fallback path: minimum_version setter fails, options are used instead
    s = SocketTransport("example.com", 1, verify=True)
    ctx = _DummySSLContext(raise_on_minimum=True)

    with (
        mock.patch("certifi.where", return_value="/tmp/ca.pem"),
        mock.patch("ssl.create_default_context", return_value=ctx),
    ):
        result = s._build_ssl_context()
        assert result is ctx

        # minimum_version was not set
        assert ctx.minimum_version is None

        # Fallback disables older protocols when options are present
        if hasattr(ssl, "OP_NO_TLSv1"):
            assert (ctx.options & ssl.OP_NO_TLSv1) != 0
        if hasattr(ssl, "OP_NO_TLSv1_1"):
            assert (ctx.options & ssl.OP_NO_TLSv1_1) != 0


def test_utils_socket_start_tls_no_socket_raises():
    """SocketTransport() Start TLS No Socket Raises."""
    s = SocketTransport("example.com", 1, secure=True)
    with pytest.raises(AppriseSocketError):
        s.start_tls()


def test_utils_socket_start_tls_already_tls_noop():
    """SocketTransport() Start TLS Already TLS Noop."""
    s = SocketTransport("example.com", 1, secure=True)
    s._sock = _DummySocket()
    s._is_tls = True
    s.start_tls()
    assert s._sock is not None
    assert s._is_tls is True


def test_utils_socket_tls_ssl_init_errors():
    """SocketTransport() Start TLS/SSL Initialization Errors."""
    s = SocketTransport("example.com", 1, secure=True)
    s._sock = _DummySocket()

    with mock.patch.object(s, "_build_ssl_context") as mctx:
        ctx = mock.Mock()
        ctx.wrap_socket.side_effect = ssl.SSLError("boom")
        mctx.return_value = ctx

        with pytest.raises(AppriseSocketError):
            s.start_tls()
        assert s._sock is None

    s._sock = _DummySocket()
    with mock.patch.object(s, "_build_ssl_context") as mctx:
        ctx = mock.Mock()
        ctx.wrap_socket.side_effect = OSError("boom")
        mctx.return_value = ctx

        with pytest.raises(AppriseSocketError):
            s.start_tls()
        assert s._sock is None


def test_utils_socket_tls_tests():
    """SocketTransport() TLS Coverage"""
    s = SocketTransport("example.com", 1, secure=True)
    base_sock = _DummySocket()
    s._sock = base_sock

    tls_sock = _DummySocket()
    ctx = mock.Mock()
    ctx.wrap_socket.return_value = tls_sock

    with (
        mock.patch.object(s, "_build_ssl_context", return_value=ctx),
        mock.patch.object(
            s, "_server_hostname_for_tls", return_value="example.com"
        ),
    ):
        s.start_tls()

    assert s._sock is tls_sock
    assert s.is_tls is True
    assert s.local_addr == ("127.0.0.1", 12345)
    assert s.remote_addr == ("203.0.113.10", 6667)
    assert s._rfile is not None
    assert s._wfile is not None


def test_utils_socket_connect():
    """SocketTransport() Connect() Coverage."""
    s = SocketTransport("example.com", 6667, secure=False, timeout=(1.0, 2.0))

    dummy = _DummySocket()

    with mock.patch("socket.socket", return_value=dummy):
        s.connect()
        assert s.connected is True
        assert s.is_tls is False
        assert s.local_addr == ("127.0.0.1", 12345)
        assert s.remote_addr == ("203.0.113.10", 6667)

    # Failure path: connect raises
    s = SocketTransport("example.com", 6667, secure=False, timeout=(1.0, 2.0))
    dummy = _DummySocket()
    dummy.connect = mock.Mock(side_effect=OSError("fail"))

    with mock.patch("socket.socket", return_value=dummy):
        with pytest.raises(AppriseSocketError):
            s.connect()
        assert s.connected is False


def test_utils_socket_read_nonblocking():
    """SocketTransport() Read Nonblocking."""
    s = SocketTransport("example.com", 1)
    sock = _DummySocket()
    s._sock = sock

    # Non-blocking success
    assert s.read(blocking=False) == b"data"

    # Non-blocking no data
    sock._recv_side_effect = BlockingIOError()
    assert s.read(blocking=False) == b""

    # Non-blocking OSError -> close + raise
    sock._recv_side_effect = OSError("boom")
    with pytest.raises(AppriseSocketError):
        s.read(blocking=False)
    assert s._sock is None


def test_utils_socket_read_blocking_timeout():
    """SocketTransport() Read Blocking Timeout."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.1))
    sock = _DummySocket()
    s._sock = sock

    # Timeout path: can_read returns False => b""
    with mock.patch.object(s, "can_read", return_value=False):
        assert s.read(blocking=True) == b""

    # Indefinite path: wait_timeout=None, loop until can_read True
    s = SocketTransport("example.com", 1, timeout=None)
    sock = _DummySocket()
    s._sock = sock

    side_effect = [False, False, True]
    with mock.patch.object(s, "can_read", side_effect=side_effect):
        assert s.read(blocking=True) == b"data"


def test_utils_socket_read_blocking_edge_cases() -> None:
    """SocketTransport() Read Blocking Edge Cases"""
    s = SocketTransport("example.com", 1, timeout=None)

    bad_sock = _DummySocket()
    good_sock = _DummySocket()
    good_sock.recv = mock.Mock(return_value=b"data")

    s._sock = bad_sock
    s._had_io = True

    def _close_side_effect() -> None:
        # emulate closing the socket
        if s._sock is not None:
            s._sock = None
        s._rfile = None
        s._wfile = None
        s._is_tls = False
        s.local_addr = None
        s.remote_addr = None

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    with (
        mock.patch.object(s, "close", side_effect=_close_side_effect),
        mock.patch.object(s, "connect", side_effect=_connect_side_effect),
        mock.patch.object(s, "can_read", return_value=None),
    ):
        # can_read(None) triggers reconnect branch, then recv() on new socket
        assert s.read(blocking=True, retries=1) == b"data"


def test_utils_socket_write_with_no_socket():
    """SocketTransport() write() - No Socket."""
    s = SocketTransport("example.com", 1)
    with pytest.raises(AppriseSocketError):
        s.write(b"hi")

    s._sock = _DummySocket()
    with pytest.raises(AppriseInvalidData):
        s.write("hi")  # type: ignore[arg-type]


def test_utils_socket_write_flush():
    """SocketTransport() write() - Flush."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.2))
    sock = _DummySocket()
    s._sock = sock
    s._wfile = _DummyFile()

    # Normal send, flush=True
    with mock.patch.object(s, "can_write", return_value=True):
        assert s.write(b"test", flush=True) == 4
    assert s._wfile.flushed >= 1

    # send returning 0 triggers connection lost
    sock = _DummySocket()
    sock.send = mock.Mock(return_value=0)
    s._sock = sock
    with pytest.raises(AppriseSocketError):
        s.write(b"test", timeout=0.1)

    # Timeout path: can_write returns False
    sock = _DummySocket()
    s._sock = sock
    with (
        mock.patch.object(s, "can_write", return_value=False),
        pytest.raises(AppriseSocketError),
    ):
        s.write(b"test", timeout=0.1)

    # OSError triggers close
    sock = _DummySocket()
    sock._send_side_effect = OSError("boom")
    s._sock = sock
    with pytest.raises(AppriseSocketError):
        s.write(b"test", timeout=0.1)
    assert s._sock is None


def test_utils_socket_read_timeouts():
    """SocketTransport() Read Timeouts"""
    # Negative connect timeout
    with pytest.raises(AppriseInvalidData):
        SocketTransport("example.com", 1, timeout=(-0.1, 1.0))

    # Negative read timeout
    with pytest.raises(AppriseInvalidData):
        SocketTransport("example.com", 1, timeout=(1.0, -0.1))

    # Force the read_t is None branch
    s = SocketTransport("example.com", 1, timeout=(1.0, None))
    assert s._connect_timeout == pytest.approx(1.0)
    assert s._read_timeout is None


def test_utils_socket_exceptions():
    """SocketTransport() Exception Testing."""
    assert issubclass(AppriseSocketError, AppriseException)


def test_utils_socket_close_exceptions():
    """SocketTransport() Close Exceptions."""

    class _BadShutdownSocket(_DummySocket):
        def shutdown(self, *_args, **_kwargs) -> None:
            raise Exception("shutdown fail")

    class _BadFile(_DummyFile):
        def flush(self) -> None:
            raise Exception("flush fail")

        def close(self) -> None:
            raise Exception("close fail")

    s = SocketTransport("example.com", 1)
    s._sock = _BadShutdownSocket()
    s._wfile = _BadFile()
    s._rfile = _BadFile()

    # Should not raise, despite flush/close/shutdown throwing
    s.close()

    assert s._sock is None
    assert s._wfile is None
    assert s._rfile is None


def test_utils_socket_refresh():
    """SocketTransport() Refresh."""
    s = SocketTransport("example.com", 1)
    s._sock = None
    s._rfile = _DummyFile()
    s._wfile = _DummyFile()

    s._refresh_wrappers()

    assert s._rfile is None
    assert s._wfile is None


def test_utils_socket_can_read_returns_bool():
    """SocketTransport() Can Read Returns Bool."""
    s = SocketTransport("example.com", 1)
    s._sock = _DummySocket()

    with mock.patch("select.select", return_value=([s._sock], [], [])):
        assert s.can_read(0.01) is True

    with mock.patch("select.select", return_value=([], [], [])):
        assert s.can_read(0.01) is False


def test_utils_socket_connect_bind():
    """SocketTransport() bind() tests"""
    s = SocketTransport(
        "example.com", 6667, bind_addr="127.0.0.1", bind_port=0
    )
    dummy = _DummySocket()
    dummy.bind = mock.Mock()

    with mock.patch("socket.socket", return_value=dummy):
        s.connect()

    dummy.bind.assert_called_once()


def test_utils_socket_connect_settimeout_handling():
    """SocketTransport() Connect settimeout() handling."""
    s = SocketTransport("example.com", 6667, timeout=None)
    dummy = _DummySocket()
    dummy.settimeout = mock.Mock()

    with mock.patch("socket.socket", return_value=dummy):
        s.connect()

    # called once with None after connect() (sock.settimeout(None))
    # but NOT called with a float connect timeout prior to connect.
    assert dummy.settimeout.call_args_list == [mock.call(None)]


def test_utils_socket_secure_connect():
    """SocketTransport() Secure connect()."""
    s = SocketTransport("example.com", 6667, secure=True)
    dummy = _DummySocket()

    with (
        mock.patch("socket.socket", return_value=dummy),
        mock.patch.object(s, "start_tls", autospec=True) as m,
    ):
        s.connect()
        m.assert_called_once()


def test_utils_socket_connect_exceptions():
    """SocketTransport() Connect Exceptions."""

    class _BadCloseSocket(_DummySocket):
        def connect(self, *_args, **_kwargs) -> None:
            raise OSError("connect fail")

        def close(self) -> None:
            raise Exception("close fail")

    s = SocketTransport("example.com", 6667, secure=False)
    dummy = _BadCloseSocket()

    with (
        mock.patch("socket.socket", return_value=dummy),
        pytest.raises(AppriseSocketError),
    ):
        s.connect()

    # transport should not keep the socket
    assert s.connected is False


def test_utils_socket_write_deadline():
    """SocketTransport() Write Deadline."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.2))
    sock = _DummySocket()
    s._sock = sock

    with mock.patch.object(s, "can_write", return_value=True) as m_can_write:
        written = s.write(b"test", timeout=0.1, flush=False)
        assert written == 4
        assert m_can_write.called


def test_utils_socket_write_timeout():
    """SocketTransport() write() timeout."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.2))
    s._sock = _DummySocket()

    # First monotonic call sets deadline, second call makes remaining <= 0
    with mock.patch("time.monotonic", side_effect=[1000.0, 1000.5]):
        with pytest.raises(AppriseSocketError) as e:
            s.write(b"test", timeout=0.1, flush=False)
        assert "Timed out during write" in str(e.value)

    s = SocketTransport("example.com", 1, timeout=None)
    sock = _DummySocket()
    s._sock = sock
    s._wfile = _DummyFile()

    # If deadline is None, can_write should never be called
    with mock.patch.object(s, "can_write") as m_can_write:
        result = s.write(b"test", flush=True, timeout=None)
        assert result == 4
        m_can_write.assert_not_called()


def test_utils_socket_write_flush_edge_cases():
    """SocketTransport() write() flush() edge cases."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.2))
    s._sock = _DummySocket()
    s._wfile = _DummyFile()

    with mock.patch.object(s._wfile, "flush") as m:
        with mock.patch.object(s, "can_write", return_value=True):
            written = s.write(b"test", flush=False, timeout=0.1)
            assert written == 4
        m.assert_not_called()

    s = SocketTransport("example.com", 1, timeout=(1.0, 0.2))
    s._sock = _DummySocket()
    s._wfile = None

    # Should succeed and simply skip flush
    with mock.patch.object(s, "can_write", return_value=True):
        assert s.write(b"test", flush=True, timeout=0.1) == 4


def test_utils_socket_recv_error():
    """SocketTransport() recv() errors

    Covers blocking read() OSError handling after readiness check.
    """
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    sock = _DummySocket()
    sock._recv_side_effect = OSError("recv failed")
    s._sock = sock

    with (
        mock.patch.object(s, "can_read", return_value=True),
        pytest.raises(AppriseSocketError) as e,
    ):
        s.read(blocking=True)

    assert "recv failed" in str(e.value)
    assert s._sock is None


def test_utils_socket_write_empty_payload_does_not_set_had_io() -> None:
    """SocketTransport() Write Empty Payload Does Not Set Had Io.

    Covers the branch where total_sent == 0, so _had_io is not updated.
    """
    s = SocketTransport("example.com", 1, timeout=None)
    s._sock = _DummySocket()
    s._wfile = _DummyFile()

    # Empty payload performs no send() calls
    sent = s.write(b"", flush=False, timeout=None)
    assert sent == 0
    assert s._had_io is False


def test_utils_socket_write_handling() -> None:
    """SocketTransport() Write handling"""
    s = SocketTransport("example.com", 1, timeout=None)
    s._sock = _DummySocket()

    # Mark prior I/O so reconnect is considered eligible in real logic, but we
    # patch _attempt_reconnect to force the continue path deterministically.
    s._had_io = True

    # Force the send to fail with an AppriseSocketError
    s._sock.send = mock.Mock(side_effect=AppriseSocketError("boom"))

    # retries=0 => attempts == 1; we force _attempt_reconnect to return True,
    # which triggers `continue` and immediately exhausts the loop.
    with (
        mock.patch.object(s, "_attempt_reconnect", return_value=True),
        pytest.raises(AppriseSocketError) as e,
    ):
        s.write(b"test", timeout=None, retries=0)

    assert "Socket write failed" in str(e.value)


def test_utils_socket_write_reconnect_continue_path_is_reached() -> None:
    """SocketTransport() Write Reconnect Continue Path Is Reached.

    Covers the `continue` path in write() when _attempt_reconnect returns True.
    Trigger AppriseSocketError (not OSError) so we do not reset _had_io via.
    close() before reconnect eligibility is checked.
    """
    s = SocketTransport("example.com", 1, timeout=None)

    bad_sock = _DummySocket()
    # send returning 0 triggers "Connection lost during write"
    bad_sock.send = mock.Mock(return_value=0)

    good_sock = _DummySocket()
    good_sock.send = mock.Mock(return_value=4)

    s._sock = bad_sock
    s._had_io = True

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    with mock.patch.object(s, "connect", side_effect=_connect_side_effect):
        assert s.write(b"test", timeout=None, retries=1, flush=False) == 4


def test_utils_socket_read_reconnect_continue_path_is_reached() -> None:
    """SocketTransport() Read Reconnect Continue Path Is Reached.

    Covers the `continue` path in read() when _attempt_reconnect returns True.
    Trigger AppriseSocketError (not OSError) so we do not reset _had_io via.
    close() before reconnect eligibility is checked.
    """
    s = SocketTransport("example.com", 1, timeout=None)

    bad_sock = _DummySocket()
    # recv returning b"" triggers "Connection lost during read"
    bad_sock.recv = mock.Mock(return_value=b"")

    good_sock = _DummySocket()
    good_sock.recv = mock.Mock(return_value=b"data")

    s._sock = bad_sock
    s._had_io = True

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    with mock.patch.object(s, "connect", side_effect=_connect_side_effect):
        assert s.read(blocking=False, retries=1) == b"data"


def test_utils_socket_attempt_reconnect_retries_zero_returns_false() -> None:
    """SocketTransport() Attempt Reconnect Retries Zero Returns False.

    Covers _attempt_reconnect() early return when retries <= 0.
    """
    s = SocketTransport("example.com", 1, timeout=None)
    s._had_io = True
    assert (
        s._attempt_reconnect(
            retries=0,
            action="read",
            exc=Exception("boom"),
        )
        is False
    )


def test_utils_socket_read_blocking_connection() -> None:
    """SocketTransport() Read Blocking Connection."""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    sock = _DummySocket()
    sock.recv = mock.Mock(return_value=b"")
    s._sock = sock

    # Mark prior I/O, but retries=0 means reconnect is not allowed anyway
    s._had_io = True

    with (
        mock.patch.object(s, "can_read", return_value=True),
        pytest.raises(AppriseSocketError) as e,
    ):
        s.read(blocking=True, retries=0)

    assert "Connection lost during read" in str(e.value)


def test_utils_socket_read_blocking() -> None:
    """SocketTransport() Read Blocking"""
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))

    bad_sock = _DummySocket()
    bad_sock.recv = mock.Mock(return_value=b"")
    good_sock = _DummySocket()
    good_sock.recv = mock.Mock(return_value=b"data")

    s._sock = bad_sock
    s._had_io = True

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    with (
        mock.patch.object(s, "can_read", return_value=True),
        mock.patch.object(s, "connect", side_effect=_connect_side_effect),
    ):
        assert s.read(blocking=True, retries=1) == b"data"

    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    sock = _DummySocket()
    s._sock = sock

    with mock.patch.object(s, "can_read", return_value=True):
        assert s.read(blocking=True) == b"data"

    s = SocketTransport("example.com", 1, timeout=None)
    s._sock = _DummySocket()

    with mock.patch.object(s, "can_read", return_value=None):
        with pytest.raises(AppriseSocketError) as e:
            s.read(blocking=True)
        assert "Socket closed" in str(e.value)

    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    s._sock = _DummySocket()

    with mock.patch.object(s, "can_read", return_value=False):
        assert s.read(blocking=True) == b""


def test_utils_socket_read_edge_cases():
    """SocketTransport() read() edge case tests."""
    s = SocketTransport("example.com", 1)
    s._sock = None
    assert s.read() == b""

    s = SocketTransport("example.com", 1, timeout=None)

    sock = _DummySocket()
    sock.recv = mock.Mock(return_value=b"")  # triggers AppriseSocketError path
    s._sock = sock
    s._had_io = True

    with (
        mock.patch.object(s, "_attempt_reconnect", return_value=True),
        pytest.raises(AppriseSocketError) as e,
    ):
        # retries=0 => attempts starts at 1, decremented to 0 in-loop
        s.read(blocking=False, retries=0)

    assert "Socket read failed" in str(e.value)


def test_utils_socket_attempt_reconnect_no_io() -> None:
    """SocketTransport() _attempt_reconnect() returns False"""
    s = SocketTransport("example.com", 1, timeout=None)

    # Explicitly not eligible for reconnect
    s._had_io = False

    with (
        mock.patch.object(s, "close") as m_close,
        mock.patch.object(s, "connect") as m_connect,
    ):
        assert (
            s._attempt_reconnect(
                retries=1,
                action="read",
                exc=Exception("boom"),
            )
            is False
        )

        # Should not attempt any reconnect work
        m_close.assert_not_called()
        m_connect.assert_not_called()


def test_utils_socket_attempt_reconnect_connect_exception() -> None:
    """SocketTransport() _attempt_reconnect() handles exception."""
    s = SocketTransport("example.com", 1, timeout=None)

    # Eligible for reconnect
    s._had_io = True

    with (
        mock.patch.object(s, "close") as m_close,
        mock.patch.object(
            s, "connect", side_effect=Exception("nope")
        ) as m_connect,
    ):
        assert (
            s._attempt_reconnect(
                retries=1,
                action="write",
                exc=Exception("boom"),
            )
            is False
        )

        m_close.assert_called_once()
        m_connect.assert_called_once()


def test_utils_socket_ssl_read_blocking() -> None:
    """SocketTransport() read() blocking"""
    s = SocketTransport("example.com", 1, timeout=None)
    sock = _DummySocket()

    # First recv indicates TLS wants more reads, second returns data
    sock.recv = mock.Mock(side_effect=[ssl.SSLWantReadError(), b"data"])
    s._sock = sock

    # Indefinite blocking, so we must reach recv() and then loop on WANT_READ
    with mock.patch.object(s, "can_read", return_value=True):
        assert s.read(blocking=True, timeout=None) == b"data"

    # Coverage for read blocking, WANT_WRITE
    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    sock = _DummySocket()

    # Trigger the WANT_WRITE/BlockingIOError handling inside the recv loop
    sock.recv = mock.Mock(side_effect=ssl.SSLWantWriteError())
    s._sock = sock

    # 1) initial readiness check returns True (enter recv loop)
    # 2) inner loop calls can_read(min(...)) and gets False, returning b""
    with mock.patch.object(s, "can_read", side_effect=[True, False]):
        assert s.read(blocking=True, timeout=0.5) == b""

    s = SocketTransport("example.com", 1, timeout=(1.0, 0.5))
    sock = _DummySocket()

    # WANT_WRITE first, then a successful read
    sock.recv = mock.Mock(side_effect=[ssl.SSLWantWriteError(), b"data"])
    s._sock = sock

    # can_read used twice:
    # - readiness check (True)
    # - busy-loop avoidance check (True)
    with mock.patch.object(s, "can_read", side_effect=[True, True]):
        assert s.read(blocking=True, timeout=0.5) == b"data"

    s = SocketTransport("example.com", 1, timeout=None)
    bad_sock = _DummySocket()

    # In blocking read, b"" triggers "Connection lost during read"
    bad_sock.recv = mock.Mock(return_value=b"")

    good_sock = _DummySocket()
    # This targets the post-reconnect immediate recv path and forces b"" there
    good_sock.recv = mock.Mock(return_value=b"")

    s._sock = bad_sock
    s._had_io = True

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    # Ensure we enter the blocking path and do the initial recv that fails,
    # then reconnect, then attempt immediate recv on the new socket.
    with (
        mock.patch.object(s, "can_read", return_value=True),
        mock.patch.object(s, "connect", side_effect=_connect_side_effect),
    ):
        with pytest.raises(AppriseSocketError) as e:
            s.read(blocking=True, retries=1)
        assert "Connection lost during read" in str(e.value)

    s = SocketTransport("example.com", 1, timeout=None)

    bad_sock = _DummySocket()
    bad_sock.recv = mock.Mock(return_value=b"")
    s._sock = bad_sock
    s._had_io = True

    good_sock = _DummySocket()
    # - First call is the post-reconnect immediate recv: it raises
    #   BlockingIOError
    # - Second call occurs in the next loop iteration after can_read says True
    good_sock.recv = mock.Mock(side_effect=[BlockingIOError(), b"data"])

    def _connect_side_effect() -> None:
        s._sock = good_sock
        s._refresh_wrappers()

    # can_read must return True so we get to recv() on the next iteration
    with (
        mock.patch.object(s, "can_read", return_value=True),
        mock.patch.object(s, "connect", side_effect=_connect_side_effect),
    ):
        assert s.read(blocking=True, retries=1) == b"data"
